#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 13 08:26:50 2018

@author: prmiles
"""

from pymcmcstat.samplers.Adaptation import cholupdate, initialize_covariance_mean_sum
from pymcmcstat.samplers.Adaptation import Adaptation
from pymcmcstat.samplers.Adaptation import is_semi_pos_def_chol, update_covariance_mean_sum
from pymcmcstat.samplers.Adaptation import unpack_simulation_options, unpack_covariance_settings
from pymcmcstat.samplers.Adaptation import below_burnin_threshold
from pymcmcstat.samplers.Adaptation import update_delayed_rejection
from pymcmcstat.samplers.Adaptation import update_cov_via_ram
from pymcmcstat.samplers.Adaptation import scale_cholesky_decomposition
from pymcmcstat.samplers.Adaptation import adjust_cov_matrix
from pymcmcstat.samplers.Adaptation import check_for_singular_cov_matrix
from pymcmcstat.samplers.Adaptation import update_cov_from_covchain
from pymcmcstat.samplers.Adaptation import setup_w_R, setup_cholupdate
from pymcmcstat.settings.SimulationOptions import SimulationOptions
from pymcmcstat.procedures.CovarianceProcedures import CovarianceProcedures
import test.general_functions as gf
import unittest
from mock import patch
import numpy as np
import io
import sys

def setup_options(**kwargs):
    SO = SimulationOptions()
    SO.define_simulation_options(**kwargs)
    return SO

def setup_run_adapt_settings():
    __, options, parameters, data = gf.setup_mcmc()
    CP = CovarianceProcedures()
    CP._initialize_covariance_settings(parameters = parameters, options = options)
    rejected = {'in_adaptation_interval': 4, 'total': 10, 'outside_bounds': 1}
    isimu = 100
    iiadapt = 100
    chain = np.zeros([100,2])
    chain[:,0] = np.linspace(1,100,100)
    chain[:,1] = np.linspace(1,100,100)
    chainind = 100
    u = np.random.random_sample(size = (1,2))
    npar = 2
    alpha = 0.78
    return CP, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha

# -------------------------------------------
class RunAdaptation(unittest.TestCase):
    def test_run_adapt_isimu_lt_burnintime(self):
        covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha = setup_run_adapt_settings()
        options.burnintime = 1000
        A = Adaptation()
        
        rejected['in_adaptation_interval'] =  96
        tstR = np.array([[0.1, 0.],[0., 0.25]])
        covariance._R = tstR.copy()
        options.burnin_scale = 0.5
        options.verbosity = 10
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        cout = A.run_adaptation(covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha)
        sys.stdout = sys.__stdout__

        self.assertEqual(capturedOutput.getvalue(), ' (burnin/down) 96.0\n', msg = 'Expected string {}'.format(' (burnin/down) 96.0\n'))
        self.assertTrue(np.isclose(cout._R, tstR/options.burnin_scale).all(), msg = str('Expect arrays to match within numerical precision: {} new {}'.format(cout._R, tstR/options.burnin_scale)))
        
    def test_run_adapt_update_covariance_mean_sum_R_is_not_none(self):
        covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha = setup_run_adapt_settings()
        
        oldcov = np.array([[0.5, 0.1],[0.1, 0.3]])
        oldmean = np.array([10.2, 2.4])
        oldwsum = 100*np.ones([1])
        oldR = np.array([[0.3, 0.1],[0, 0.4]])
        
        covariance._R = oldR.copy()
        covariance._wsum = oldwsum.copy()
        covariance._meanchain = oldmean.copy()
        covariance._covchain = oldcov.copy()
        
        A = Adaptation()
        cout = A.run_adaptation(covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha)
        
        tsmtx = np.array([[827.030150753768,	905.811055276382],[905.811055276382,	1000.17688442211]])
        tsRmtx = np.array([[48.8041682048865, 53.4531359748650],[0.,	4.82407313255805]])
        tsmean = np.array([30.3500000000000,	26.4500000000000])
        self.assertTrue(np.isclose(tsmtx, cout._covchain).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, cout._covchain)))
        self.assertTrue(np.isclose(tsmean, cout._meanchain).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, cout._meanchain)))
        self.assertTrue(np.isclose(tsRmtx, cout._R).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsRmtx, cout._R)))
        
    @patch('pymcmcstat.samplers.Adaptation.update_cov_via_ram', return_value = np.array([[0.26917331, 0.09473752],[0.09473752, 0.49657009]]))
    def test_run_adapt_update_covariance_mean_sum_doram_is_true(self, mock_ram):
        covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha = setup_run_adapt_settings()
        
        options.doram = True
        
        oldcov = np.array([[0.5, 0.1],[0.1, 0.3]])
        oldmean = np.array([10.2, 2.4])
        oldwsum = 100*np.ones([1])
        oldR = np.array([[0.3, 0.1],[0, 0.4]])
        
        covariance._R = oldR.copy()
        covariance._wsum = oldwsum.copy()
        covariance._meanchain = oldmean.copy()
        covariance._covchain = oldcov.copy()
        
        A = Adaptation()
        cout = A.run_adaptation(covariance, options, isimu, iiadapt, rejected, chain, chainind, u, npar, alpha)
        
        tsmtx = np.array([[827.030150753768,	905.811055276382],[905.811055276382,	1000.17688442211]])
        tsRmtx = np.array([[0.880465293353463,	0.309886215458656],[0.,	1.15502917394701]])
        tsmean = np.array([30.3500000000000,	26.4500000000000])
        self.assertTrue(np.isclose(tsmtx, cout._covchain).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, cout._covchain)))
        self.assertTrue(np.isclose(tsmean, cout._meanchain).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, cout._meanchain)))
        self.assertTrue(np.isclose(tsRmtx, cout._R).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsRmtx, cout._R)))
   
# --------------------------------------------
class CholUpdate(unittest.TestCase):
    def test_cholupate(self):
        R = np.diag([1., 1., 1.])
        R1 = cholupdate(R = R, x = np.array([1,1,1], dtype = float))
        tsmtx = np.array([[1.41421356237310,	0.707106781186548, 0.707106781186548],[0,	1.22474487139159,	0.408248290463863],[0,	0,	1.15470053837925]])
        self.assertTrue(np.isclose(R1, tsmtx).all(), msg = str('Expect arrays to match within numerical precision: {} new {}'.format(R1, tsmtx)))

# --------------------------------------------
class SetupWR(unittest.TestCase):
    def test_setup_w_R_none_input(self):
        w = None
        oldR = None
        n = 100
        wout, Rout = setup_w_R(w = w, oldR = oldR, n = n)
        self.assertTrue(np.array_equal(wout, np.ones(n)*np.ones(1)), msg = 'Expect np.ones([n])')
        self.assertEqual(Rout, None, msg = 'Expect None')
        
    def test_setup_w_R_not_none_input(self):
        w = 100*np.ones(1)
        oldR = np.random.random_sample(size = (3,3))
        n = 100
        wout, Rout = setup_w_R(w = w, oldR = oldR, n = n)
        self.assertTrue(np.array_equal(wout, 100*np.ones(n)*np.ones(1)), msg = 'Expect 100*np.ones([n])')
        self.assertTrue(np.array_equal(Rout, oldR), msg = 'Expect arrays to match')

# --------------------------------------------
class SetupCholUpdate(unittest.TestCase):
    def test_setup_cholupdate(self):
        x = np.zeros([100,2])
        x[:,0] = np.linspace(1,100,100)
        x[:,1] = np.linspace(1,100,100)
        xi = x[0,:]
        wsum = np.ones([1])
        oldmean = np.array([10.2, 2.4])
        oldwsum = 100*np.ones([1])
        R = np.array([[0.3, 0.1],[0, 0.4]])
        Rin, xin = setup_cholupdate(R = R, oldwsum = oldwsum, wsum = wsum, oldmean = oldmean, xi = xi)
        
        tstRin = np.array([[0.298496231131986, 0.0994987437106620],[0, 0.397994974842648]])
        tstxin = np.array([-0.915434214993190,-0.139305206629398])
        self.assertTrue(np.isclose(Rin, tstRin).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(Rin, tstRin)))
        self.assertTrue(np.isclose(xin, tstxin).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(xin, tstxin)))

# --------------------------------------------
class UpdateCovarianceMeanSum(unittest.TestCase):

    def test_update_covariance_mean_sum_none(self):
        x = np.array([[]])
        w = np.ones([1])
        xcov, xmean, wsum = update_covariance_mean_sum(x = x, w = w, oldcov = None, oldmean = None, oldwsum = None, oldR = None)
        self.assertEqual(xcov, None, msg = 'Expect None')
        self.assertEqual(xmean, None, msg = 'Expect None')
        self.assertEqual(wsum, None, msg = 'Expect None')
        
    def test_update_covariance_mean_sum_initialize(self):
        x = np.zeros([100,2])
        x[:,0] = np.linspace(1,100,100)
        x[:,1] = np.linspace(1,100,100)
        w = np.ones([1])
        xcov, xmean, wsum = update_covariance_mean_sum(x = x, w = w, oldcov = None, oldmean = None, oldwsum = None, oldR = None)
        tsmtx = np.array([[8.4167e+02, 8.4167e+02], [8.4167e+02,   8.4167e+02]])
        tsmean = np.array([5.0500e+01, 5.0500e+01])
        self.assertTrue(isinstance(xcov, np.ndarray), msg = 'Expect numpy array')
        self.assertEqual(xcov.shape, (2,2), msg = 'Expect shape = (2,2)')
        self.assertTrue(np.isclose(tsmtx, xcov).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, xcov)))
        self.assertTrue(np.isclose(tsmean, xmean).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, xmean)))
        
        self.assertTrue(np.isclose(xmean, np.mean(x, axis = 0)).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(xmean, np.mean(x, axis = 0))))
        self.assertEqual(wsum, 100, msg = 'Expect wsum = 100')
        
    def test_update_covariance_mean_sum_R_is_none(self):
        x = np.zeros([100,2])
        x[:,0] = np.linspace(1,100,100)
        x[:,1] = np.linspace(1,100,100)
        w = np.ones([1])
        oldcov = np.array([[0.5, 0.1],[0.1, 0.3]])
        oldmean = np.array([10.2, 2.4])
        oldwsum = 100*np.ones(1)
        xcov, xmean, wsum = update_covariance_mean_sum(x = x, w = w, oldcov = oldcov, oldmean = oldmean, oldwsum = oldwsum, oldR = None)
        tsmtx = np.array([[827.030150753768,	905.811055276382],[905.811055276382,	1000.17688442211]])
        tsmean = np.array([30.3500000000000,	26.4500000000000])
        self.assertTrue(isinstance(xcov, np.ndarray), msg = 'Expect numpy array')
        self.assertEqual(xcov.shape, (2,2), msg = 'Expect shape = (2,2)')
        self.assertTrue(np.isclose(tsmtx, xcov).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, xcov)))
        self.assertTrue(np.isclose(tsmean, xmean).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, xmean)))
        self.assertEqual(wsum, 200, msg = 'Expect wsum = 200')
        
    def test_update_covariance_mean_sum_R_is_not_none(self):
        x = np.zeros([100,2])
        x[:,0] = np.linspace(1,100,100)
        x[:,1] = np.linspace(1,100,100)
        w = np.ones([1])
        oldcov = np.array([[0.5, 0.1],[0.1, 0.3]])
        oldmean = np.array([10.2, 2.4])
        oldwsum = 100*np.ones([1])
        oldR = np.array([[0.3, 0.1],[0, 0.4]])
        xcov, xmean, wsum = update_covariance_mean_sum(x = x, w = w, oldcov = oldcov, oldmean = oldmean, oldwsum = oldwsum, oldR = oldR)
        tsmtx = np.array([[827.030150753768,	905.811055276382],[905.811055276382,	1000.17688442211]])
        tsmean = np.array([30.3500000000000,	26.4500000000000])
        self.assertTrue(isinstance(xcov, np.ndarray), msg = 'Expect numpy array')
        self.assertEqual(xcov.shape, (2,2), msg = 'Expect shape = (2,2)')
        self.assertTrue(np.isclose(tsmtx, xcov).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, xcov)))
        self.assertTrue(np.isclose(tsmean, xmean).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, xmean)))
        self.assertEqual(wsum, 200, msg = 'Expect wsum = 200')

# --------------------------------------------
class InitializeCovarianceMeanSum(unittest.TestCase):
    def test_initialize_covariance_mean_sum(self):
        x = np.zeros([100,2])
        x[:,0] = np.linspace(1,100,100)
        x[:,1] = np.linspace(1,100,100)
        w = np.ones([100,])
        xcov, xmean, wsum = initialize_covariance_mean_sum(x = x, w = w)
        tsmtx = np.array([[8.4167e+02, 8.4167e+02], [8.4167e+02,   8.4167e+02]])
        tsmean = np.array([5.0500e+01, 5.0500e+01])
        self.assertTrue(isinstance(xcov, np.ndarray), msg = 'Expect numpy array')
        self.assertEqual(xcov.shape, (2,2), msg = 'Expect shape = (2,2)')
        self.assertTrue(np.isclose(tsmtx, xcov).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmtx, xcov)))
        self.assertTrue(np.isclose(tsmean, xmean).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(tsmean, xmean)))
        
        self.assertTrue(np.isclose(xmean, np.mean(x, axis = 0)).all(), msg = str('Mean algorithms should agree: {} neq {}'.format(xmean, np.mean(x, axis = 0))))
        self.assertEqual(wsum, 100, msg = 'Expect wsum = 100')

# --------------------------------------------
class Initialization(unittest.TestCase):
    def test_init_adapt(self):
        AD = Adaptation()
        ADD = AD.__dict__
        check_fields = ['qcov', 'qcov_scale', 'R', 'qcov_original', 'invR', 'iacce', 'covchain', 'meanchain']
        for ii, cf in enumerate(check_fields):
            self.assertTrue(ADD[cf] is None, msg = str('Initialize {} to None'.format(cf)))
# --------------------------------------------
class IsSemiPositiveDefinite(unittest.TestCase):
    def test_semipositivedef(self):
        mtx = np.diag([2, 2])
        flag, c = is_semi_pos_def_chol(x = mtx)
        self.assertTrue(flag, msg = 'Expect true')
        self.assertTrue(isinstance(c, np.ndarray), msg = 'Expect numpy array')
        self.assertEqual(c.shape, (2,2), msg = 'Expect 2x2 array')
        
    def test_notsemipositivedef(self):
        mtx = np.array([[2, -1],[0, 0]])
        flag, c = is_semi_pos_def_chol(x = mtx)
        self.assertFalse(flag, msg = 'Expect false')
# --------------------------------------------        
class UnpackSimulationOptions(unittest.TestCase):
    def test_unpack_options(self):
        SO = setup_options()
        burnintime, burninscale, ntry, drscale, alphatarget, etaparam, qcov_adjust, doram, verbosity = unpack_simulation_options(options = SO)
        defaults = {'burnintime': 0, 'burninscale': 10, 'ntry': 2, 'drscale': np.array([5, 4, 3], dtype = float),
                    'alphatarget': 0.234, 'etaparam': 0.7, 'qcov_adjust': 1e-8, 'doram': 0, 'verbosity': 1}
        
        self.assertEqual(burnintime, defaults['burnintime'], msg = str('Expected {} = {}'.format('burnintime', defaults['burnintime'])))
        self.assertEqual(burninscale, defaults['burninscale'], msg = str('Expected {} = {}'.format('burninscale', defaults['burninscale'])))
        self.assertEqual(ntry, defaults['ntry'], msg = str('Expected {} = {}'.format('ntry', defaults['ntry'])))
        self.assertTrue(np.array_equal(drscale, defaults['drscale']), msg = str('Expected {} = {}'.format('drscale', defaults['drscale'])))
        self.assertEqual(alphatarget, defaults['alphatarget'], msg = str('Expected {} = {}'.format('alphatarget', defaults['alphatarget'])))
        self.assertEqual(etaparam, defaults['etaparam'], msg = str('Expected {} = {}'.format('etaparam', defaults['etaparam'])))
        self.assertEqual(qcov_adjust, defaults['qcov_adjust'], msg = str('Expected {} = {}'.format('qcov_adjust', defaults['qcov_adjust'])))
        self.assertEqual(doram, defaults['doram'], msg = str('Expected {} = {}'.format('doram', defaults['doram'])))
        self.assertEqual(verbosity, defaults['verbosity'], msg = str('Expected {} = {}'.format('verbosity', defaults['verbosity'])))
        
# --------------------------------------------        
class UnpackCovarianceSettings(unittest.TestCase):
    def test_unpack_covariance(self):
        model, options, parameters, data = gf.setup_mcmc()
        CP = CovarianceProcedures()
        CP._initialize_covariance_settings(parameters = parameters, options = options)
        
        last_index_since_adaptation, R, oldcovchain, oldmeanchain, oldwsum, no_adapt_index, qcov_scale, qcov = unpack_covariance_settings(covariance = CP)

        out = {'last_index_since_adaptation': last_index_since_adaptation, 'R': R, 'oldcovchain': oldcovchain, 'oldmeanchain': oldmeanchain,
                    'oldwsum': oldwsum, 'no_adapt_index': no_adapt_index, 'qcov_scale': qcov_scale, 'qcov': qcov}
        
        defaults = {'last_index_since_adaptation': 0, 'R': np.array([[0.1, 0.],[0., 0.25]]), 'oldcovchain': None, 'oldmeanchain': None,
                    'oldwsum': None, 'no_adapt_index': np.array([False, False]), 'qcov_scale': 2.4/np.sqrt(2), 'qcov': np.square(np.array([[0.1, 0.],[0., 0.25]]))}

        check_these = ['last_index_since_adaptation', 'oldcovchain', 'oldmeanchain', 'oldwsum']
        for item in check_these:
            self.assertEqual(out[item], defaults[item], msg = str('Expected {} = {} ? {}'.format(item, defaults[item], out[item])))
            
        check_these = ['R', 'no_adapt_index', 'qcov_scale', 'qcov']
        for item in check_these:
            self.assertTrue(np.array_equal(out[item], defaults[item]), msg = str('Expected {} = {} ? {}'.format(item, defaults[item], out[item])))
        
# --------------------------------------------
class BelowBurninThreshold(unittest.TestCase):
    def test_below_burnin_reject_gt_95(self):
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        R = np.array([[0.1, 0.],[0., 0.25]])
        burninscale = 0.5
        verbosity = 10
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = below_burnin_threshold(rejected = rejected, iiadapt = iiadapt, R = R, burnin_scale = burninscale, verbosity = verbosity)
        sys.stdout = sys.__stdout__

        self.assertEqual(capturedOutput.getvalue(), ' (burnin/down) 96.0\n', msg = 'Expected string {}'.format(' (burnin/down) 96.0\n'))
        self.assertTrue(np.array_equal(Rout, R/burninscale), msg = str('Expected R = {}'.format(R/burninscale)))
    
    def test_below_burnin_reject_lt_95(self):
        rejected = {'in_adaptation_interval': 4}
        iiadapt = 100
        R = np.array([[0.1, 0.],[0., 0.25]])
        burninscale = 0.5
        verbosity = 10
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = below_burnin_threshold(rejected = rejected, iiadapt = iiadapt, R = R, burnin_scale = burninscale, verbosity = verbosity)
        sys.stdout = sys.__stdout__

        self.assertEqual(capturedOutput.getvalue(), ' (burnin/up) 4.0\n', msg = 'Expected string {}'.format(' (burnin/up) 4.0\n'))
        self.assertTrue(np.array_equal(Rout, R*burninscale), msg = str('Expected R = {}'.format(R/burninscale)))
        
# --------------------------------------------
class UpdateDelayedRejection(unittest.TestCase):
    def test_update_dr_ntry_1(self):
        RDR, invR = update_delayed_rejection(R = None, npar = 2, ntry = 1, drscale = 0)
        self.assertEqual(RDR, None, msg = 'Expect None')
        self.assertEqual(invR, None, msg = 'Expect None')
        
    def test_update_dr_ntry_3(self):
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        ntry = 3
        drscale = np.array([5,4,3], dtype = float)
        RDR, invR = update_delayed_rejection(R = R, npar = npar, ntry = ntry, drscale = drscale)
        self.assertTrue(isinstance(RDR, list), msg = 'Expect list return')
        self.assertTrue(isinstance(invR, list), msg = 'Expect list return')
        self.assertTrue(np.array_equal(RDR[0], R), msg = 'Expect arrays to match')
        self.assertTrue(np.array_equal(invR[0], np.linalg.solve(R, np.eye(npar))), msg = 'Expect arrays to match')
        self.assertTrue(np.array_equal(RDR[1], RDR[0]*(drscale[0]**(-1))), msg = str('Expect arrays to match: {} neq {}'.format(RDR[1], RDR[0]/drscale[0])))
        self.assertTrue(np.array_equal(invR[1], invR[0]*drscale[0]), msg = str('Expect arrays to match: {} neq {}'.format(invR[1], invR[0]*drscale[0])))
        self.assertTrue(np.array_equal(RDR[2], RDR[1]*(drscale[1]**(-1))), msg = str('Expect arrays to match: {} neq {}'.format(RDR[2], RDR[1]/drscale[1])))
        self.assertTrue(np.array_equal(invR[2], invR[1]*drscale[1]), msg = str('Expect arrays to match: {} neq {}'.format(invR[2], invR[1]*drscale[1])))
        
# --------------------------------------------
class UpdateCovViaRam(unittest.TestCase):
    def test_update_via_ram(self):
        u = np.array([0.1, 0.5])
        isimu = 100
        etaparam = 0.7
        npar = 2
        alphatarget = 0.234
        alpha = 0.7
        R = np.array([[0.1, 0.],[0., 0.25]])
        upcov = update_cov_via_ram(u = u, isimu = isimu, etaparam = etaparam, npar = npar, alphatarget = alphatarget, alpha = alpha, R = R)
        
        self.assertEqual(upcov.shape, (2,2), msg = 'Expect shape = (2,2)')
        
        uu = u*(np.linalg.norm(u)**(-1))
        eta = (isimu**(etaparam))**(-1)
        ram = np.eye(npar) + eta*(min(1.0, alpha) - alphatarget)*(np.dot(uu.transpose(), uu))
        self.assertTrue(np.array_equal(upcov, np.dot(np.dot(R.transpose(),ram),R)), msg = str('Expect arrays to match: {} neq {}'.format(upcov, np.dot(np.dot(R.transpose(),ram),R))))
        
# --------------------------------------------
class ScaleCholeskyDecomposition(unittest.TestCase):
    def test_scale_chol_dec(self):
        Ra = np.random.random_sample(size = (2,2))
        qcov_scale = 0.2
        R = scale_cholesky_decomposition(Ra = Ra, qcov_scale = qcov_scale)
        self.assertTrue(np.array_equal(R, Ra*qcov_scale), msg = 'Expect arrays to match')
        
# --------------------------------------------
class AdjustCovMatrix(unittest.TestCase):
    def test_singular_cov_mat_successfully_adjusted(self):
        upcov = np.array([[0.1, 0.4],[0.4, 0.2]])
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        qcov_adjust = 0.9
        qcov_scale = 0.3
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        verbosity = 10
        
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = adjust_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        self.assertEqual(capturedOutput.getvalue(), 'adjusted covariance matrix\n', msg = 'Expected string {}'.format('adjusted covariance matrix\n'))
        tmp = upcov + np.eye(npar)*qcov_adjust
        pos_def_adjust, pRa = is_semi_pos_def_chol(tmp)
        self.assertTrue(pos_def_adjust, msg = 'Expect True')
        self.assertTrue(np.array_equal(pRa*qcov_scale, Rout), msg = str('Expect arrays to match: {} neq {}'.format(pRa*qcov_scale, Rout)))
        
    @patch('pymcmcstat.samplers.Adaptation.is_semi_pos_def_chol', return_value = (False, None))
    def test_singular_cov_mat_not_successfully_adjusted(self, mock_is):
        upcov = np.array([[0.1, 0.4],[0.4, 0.0]])
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        qcov_adjust = 0.9
        qcov_scale = 0.3
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        verbosity = 10
        
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = adjust_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        self.assertEqual(capturedOutput.getvalue(), 'covariance matrix singular, no adaptation 96.0\n', msg = 'Expected string {}'.format('covariance matrix singular, no adaptation 96.0\n'))
        self.assertTrue(np.array_equal(R, Rout), msg = str('Expect arrays to match: {} neq {}'.format(R, Rout)))
        
# --------------------------------------------
class CheckForSingularCovMatrix(unittest.TestCase):
    @patch('pymcmcstat.samplers.Adaptation.is_semi_pos_def_chol', return_value = (True, np.array([[0.1, 0.],[0., 0.25]])))
    def test_singular_cov_mat_scale_cholesky(self, mock_is):
        upcov = np.array([[0.1, 0.4],[0.4, 0.2]])
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        qcov_adjust = 0.9
        qcov_scale = 0.3
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        verbosity = 10
        
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = check_for_singular_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        self.assertTrue(np.array_equal(Rout, R*qcov_scale), msg = 'Expect arrays to match')
        
    def test_check_for_singular_cov_mat_successfully_adjusted(self):
        upcov = np.array([[0.1, 0.4],[0.4, 0.2]])
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        qcov_adjust = 0.9
        qcov_scale = 0.3
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        verbosity = 10
        
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = check_for_singular_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        self.assertEqual(capturedOutput.getvalue(), 'adjusted covariance matrix\n', msg = 'Expected string {}'.format('adjusted covariance matrix\n'))
        tmp = upcov + np.eye(npar)*qcov_adjust
        pos_def_adjust, pRa = is_semi_pos_def_chol(tmp)
        self.assertTrue(pos_def_adjust, msg = 'Expect True')
        self.assertTrue(np.array_equal(pRa*qcov_scale, Rout), msg = str('Expect arrays to match: {} neq {}'.format(pRa*qcov_scale, Rout)))
        
    @patch('pymcmcstat.samplers.Adaptation.is_semi_pos_def_chol', return_value = (False, None))
    def test_check_for_singular_cov_mat_not_successfully_adjusted(self, mock_is):
        upcov = np.array([[0.1, 0.4],[0.4, 0.2]])
        R = np.array([[0.1, 0.],[0., 0.25]])
        npar = 2
        qcov_adjust = 0.9
        qcov_scale = 0.3
        rejected = {'in_adaptation_interval': 96}
        iiadapt = 100
        verbosity = 10
        
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        Rout = check_for_singular_cov_matrix(upcov = upcov, R = R, npar = npar, qcov_adjust = qcov_adjust, qcov_scale = qcov_scale, rejected = rejected, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        self.assertEqual(capturedOutput.getvalue(), 'covariance matrix singular, no adaptation 96.0\n', msg = 'Expected string {}'.format('covariance matrix singular, no adaptation 96.0\n'))
        self.assertTrue(np.array_equal(R, Rout), msg = str('Expect arrays to match: {} neq {}'.format(R, Rout)))
        
# --------------------------------------------
class UpdateCovFromCovchain(unittest.TestCase):
    def test_update_cov_from_chain_all_adapt(self):
        covchain = np.random.random_sample(size = (3,3))
        qcov = np.random.random_sample(size = (3,3))
        no_adapt_index = np.array([False, False, False], dtype = bool)
        upcov = update_cov_from_covchain(covchain = covchain, qcov = qcov, no_adapt_index = no_adapt_index)
        
        self.assertTrue(np.array_equal(upcov, covchain), msg = str('Expect arrays to match: {} neq {}'.format(covchain, upcov)))
        
    def test_update_cov_from_chain_middle_no_adapt(self):
        covchain = np.random.random_sample(size = (3,3))
        qcov = np.random.random_sample(size = (3,3))
        no_adapt_index = np.array([False, True, False], dtype = bool)
        upcov = update_cov_from_covchain(covchain = covchain, qcov = qcov, no_adapt_index = no_adapt_index)
        covchain[no_adapt_index, :] = qcov[no_adapt_index,:]
        covchain[:,no_adapt_index] = qcov[:,no_adapt_index]
        self.assertTrue(np.array_equal(upcov, covchain), msg = str('Expect arrays to match: {} neq {}'.format(covchain, upcov)))
        
    def test_update_cov_from_chain_ends_no_adapt(self):
        covchain = np.random.random_sample(size = (3,3))
        qcov = np.random.random_sample(size = (3,3))
        no_adapt_index = np.array([True, False, True], dtype = bool)
        upcov = update_cov_from_covchain(covchain = covchain, qcov = qcov, no_adapt_index = no_adapt_index)
        covchain[no_adapt_index, :] = qcov[no_adapt_index,:]
        covchain[:,no_adapt_index] = qcov[:,no_adapt_index]
        self.assertTrue(np.array_equal(upcov, covchain), msg = str('Expect arrays to match: {} neq {}'.format(covchain, upcov)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 26 06:16:02 2018

General functions used throughout package

@author: prmiles
"""

def message(verbosity, level, printthis):
    '''
    Display message

    Args:
        * **verbosity** (:py:class:`int`): Verbosity of display output.
        * **level** (:py:class:`int`): Print level relative to verbosity.
        * **printthis** (:py:class:`str`): String to be printed.
    '''
    printed = False
    if verbosity >= level:
        print(printthis)
        printed = True
    return printed

def removekey(d, key):
    '''
    Removed elements from dictionary and return the remainder.

    Args:
        * **d** (:py:class:`dict`): Original dictionary.
        * **key** (:py:class:`str`): Keyword to be removed.

    Returns:
        * **r** (:py:class:`dict`): Updated dictionary without the keyword, value pair.
    '''
    r = dict(d)
    del r[key]
    return r


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
'''
Created on Wed Jan 17 09:08:13 2018

@author: prmiles
'''
# import required packages
import numpy as np
from datetime import datetime

class SimulationOptions:
    '''
    MCMC simulation options.

    Attributes:
       * :meth:`~define_simulation_options`
       * :meth:`~display_simulation_options`
    '''
    def __init__(self):
        # initialize simulation option variables
#        self.options = BaseSimulationOptions()
        self.description = 'simulation_options'
        self.__options_set = False

    def define_simulation_options(self, nsimu=int(1e4), adaptint = None, ntry = None, method='dram',
                 printint=None, adaptend = 0, lastadapt = 0, burnintime = 0,
                 waitbar = 1, debug = 0, qcov = None, updatesigma = False,
                 noadaptind = None, stats = 0, drscale = np.array([5, 4, 3], dtype = float),
                 adascale = None, savesize = 0, maxmem = 0, chainfile = 'chainfile',
                 s2chainfile = 's2chainfile', sschainfile = 'sschainfile', covchainfile = 'covchainfile', savedir = None,
                 save_to_bin = False, skip = 1, label = None, RDR = None, verbosity = 1, maxiter = None,
                 priorupdatestart = 0, qcov_adjust = 1e-8, burnin_scale = 10,
                 alphatarget = 0.234, etaparam = 0.7, initqcovn = None,
                 doram = None, rndseq = None, results_filename = None, save_to_json = False, save_to_txt = False,
                 json_restart_file = None):
        '''
        Define simulation options.

        Args:
            * **nsimu** (:py:class:`int`): Number of parameter samples to simulate.  Default is 1e4.
            * **adaptint** (:py:class:`int`): Number of interates between adaptation. Default is method dependent.
            * **ntry** (:py:class:`int`): Number of tries to take before rejection. Default is method dependent.
            * **method** (:py:class:`str`): Sampling method (:code:`'mh', 'am', 'dr', 'dram'`).  Default is :code:`'dram'`.
            * **printint** (:py:class:`int`): Printing interval.
            * **adaptend** (:py:class:`int`): Obsolete.
            * **lastadapt** (:py:class:`int`): Last adaptation iteration (i.e., no more adaptation beyond this point).
            * **burnintime** (:py:class:`int`):
            * **waitbar** (:py:class:`int`): Flag to use progress bar. Default is 1 -> on (otherwise -> off).
            * **debug** (:py:class:`int`): Flag to perform debug.  Default is 0 -> off.
            * **qcov** (:class:`~numpy.ndarray`): Proposal parameter covariance matrix.
            * **updatesigma** (:py:class:`bool`): Flag for updating measurement error variance. Default is 0 -> off (1 -> on).
            * **noadaptind** (:py:class:`int`): Indices not to be adapted in covariance matrix. Default is [] (untested).
            * **stats** (:py:class:`int`): Calculate convergence statistics. Default is 0 -> off (1 -> on).
            * **drscale** (:class:`~numpy.ndarray`): Reduced scale for sampling in DR algorithm. Default is [5,4,3].
            * **adascale** (:py:class:`float`): User defined covariance scale.  Default is method dependent (untested).
            * **savesize** (:py:class:`int`): Size of chain segments when saving to log files.  Default is 0.
            * **maxmem** (:py:class:`int`): Maximum memory available in mega bytes (Obsolete).
            * **chainfile** (:py:class:`str`): File name for :code:`chain` log file.
            * **sschainfile** (:py:class:`str`): File name for :code:`sschain` log file.
            * **s2chainfile** (:py:class:`str`): File name for :code:`s2chain` log file.
            * **covchainfile** (:py:class:`str`): File name for :code:`qcov` log file.
            * **savedir** (:py:class:`str`): Output directory of log files.  Default is current directory.
            * **save_to_bin** (:py:class:`bool`): Save log files to binary.  Default is False.
            * **save_to_txt** (:py:class:`bool`): Save log files to text.  Default is False.
            * **skip** (:py:class:`int`):
            * **label** (:py:class:`str`):
            * **RDR** (:class:`~numpy.ndarray`): R matrix for each stage of DR.
            * **verbosity** (:py:class:`int`): Verbosity of display output.
            * **maxiter** (:py:class:`int`): Obsolete.
            * **priorupdatestart**
            * **qcov_adjust** (:py:class:`float`): Adjustment scale for covariance matrix.
            * **burnin_scale** (:py:class:`float`): Scale for burnin.
            * **alphatarget** (:py:class:`float`): Acceptance ratio target.
            * **etaparam** (:py:class:`float`):
            * **initqcovn** (:py:class:`float`): Proposal covariance weight in update.
            * **doram** (:py:class:`int`): Flag to perform :code:`'ram'` algorithm (Obsolete).
            * **rndseq** (:class:`~numpy.ndarray`): Random number sequence (Obsolete).
            * **results_filename** (:py:class:`str`): Output file name when saving results structure with json.
            * **save_to_json** (:py:class:`bool`): Save results structure to json file.  Default is False.
            * **json_restart_file** (:py:class:`str`): Extract parameter covariance and last sample value from saved json file.

        .. note::

            For the log file names :code:`chainfile, sschainfile, s2chainfile` and :code:`covchainfile` do not include the extension.
            By specifying whether to save to text or to binary, the appropriate extension will be added.
        '''
        
        method_dictionary = {
            'mh': {'adaptint': 0, 'ntry': 1, 'doram': 0, 'adascale': adascale},
            'am': {'adaptint': 100, 'ntry': 1, 'doram': 0, 'adascale': adascale},
            'dr': {'adaptint': 0, 'ntry': 2, 'doram': 0, 'adascale': adascale},
            'dram': {'adaptint': 100, 'ntry': 2, 'doram': 0, 'adascale': adascale},
            'ram': {'adaptint': 1, 'ntry': 1, 'doram': 1, 'adascale': 1.},
            }
        
        # define items from dictionary
        if adaptint is None:
            self.adaptint = method_dictionary[method]['adaptint']  # update interval for adaptation
        elif method == 'mh' or method == 'dr':
            self.adaptint = method_dictionary[method]['adaptint']  # no adaptation - enforce!
        else:
            self.adaptint = adaptint
        
        if ntry is None:
            self.ntry = method_dictionary[method]['ntry']
        else:
            self.ntry = ntry
            
        if adascale is None:
            self.adascale = method_dictionary[method]['adascale']  # qcov_scale
        else:
            self.adascale = adascale
            
        if doram is None:
            self.doram = method_dictionary[method]['doram']
        else:
            self.doram = doram
        
        
        self.nsimu = nsimu  # length of chain to simulate
        self.method = method
        self.dodram = 0
        
        self.printint = printint  # print interval
        self.adaptend = adaptend  # last adapt
        self.lastadapt = lastadapt # last adapt
        self.burnintime = burnintime
        self.waitbar = waitbar # use waitbar
        self.debug = debug  # show some debug information
        self.qcov = qcov  # proposal covariance
        self.initqcovn = initqcovn  # proposal covariance weight in update
        self.updatesigma = updatesigma  #
        if noadaptind is None:
            self.noadaptind = [] # do not adapt these indices
        self.priorupdatestart = priorupdatestart
        self.qcov_adjust = qcov_adjust  # eps adjustment
        self.burnin_scale = burnin_scale
        self.alphatarget = alphatarget  # acceptance ratio target
        self.etaparam = etaparam  #
        self.stats = stats  # convergence statistics
        self.drscale = drscale

        self.skip = skip
        
        datestr = datetime.now().strftime('%Y%m%d_%H%M%S')
        self.datestr = datestr
        
        if label is None:
            self.label = str('MCMC run at {}'.format(datestr))
        else:
            self.label = label
            
        self.RDR = RDR
        self.verbosity = verbosity # amount of information to print
        self.maxiter = maxiter
        
        # log settings
        self.savesize = savesize
        self.maxmem = maxmem
            
        self.chainfile = chainfile
        self.s2chainfile = s2chainfile
        self.sschainfile = sschainfile
        self.covchainfile = covchainfile
        
        if savedir is None:
            self.savedir = str('{}_{}'.format(datestr,'chain_log'))
        else:
            self.savedir = savedir
            
        self.save_to_bin = save_to_bin
        self.save_to_txt = save_to_txt
        
        self.results_filename = results_filename
        self.save_to_json = save_to_json
        self.json_restart_file = json_restart_file
        
        self.__options_set = True # options have been defined
        
    def _check_dependent_simulation_options(self, model):
        '''
        Check dependent parameters.
        - Checks that :code:`savesize` is between 0 and :code:`nsimu`.
        - If :code:`ntry` greater than 1, then assume delayed rejection was wanted.
        - If :code:`lastadapt` less than 1, then set equal to :code:`nsimu`.
        - Update :code:`printint` based on size of :code:`adaptint`.
        - If :code:`N0` not None, turn on :code:`updatesigma`.

        Args:
            * **model**: (:class:`~.ModelSettings`): MCMC model settings.
        '''
        # save options
        if self.savesize <= 0 or self.savesize > self.nsimu:
            self.savesize = self.nsimu
        
        # turn on DR if ntry > 1
        if self.ntry > 1:
            self.dodram = 1
        else:
            self.dodram = 0
            
        if self.lastadapt < 1:
            self.lastadapt = self.nsimu
            
        if self.printint is None:
            self.printint = max(100,min(1000,self.adaptint))
            
        # if N0 given, then also turn on updatesigma
        if model.N0 is not None:
            self.updatesigma = True
            
    def display_simulation_options(self, print_these = None):
        '''
        Display subset of the simulation options.
        
        Args:
            * **print_these** (:py:class:`list`): List of strings corresponding to keywords.  Default below.

        ::

            print_these = ['nsimu', 'adaptint', 'ntry', 'method', 'printint', 'lastadapt', 'drscale', 'qcov']
        '''
        if print_these is None:
            print_these = ['nsimu', 'adaptint', 'ntry', 'method', 'printint', 'lastadapt', 'drscale', 'qcov']
            
        print('simulation options:')
        for ptii in print_these:
            print('\t{} = {}'.format(ptii, getattr(self, ptii)))
            
        return print_these


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 29 15:35:08 2018

@author: prmiles
"""
import numpy as np
from pymcmcstat.MCMC import MCMC
from pymcmcstat.settings.DataStructure import DataStructure
import os

def removekey(d, key):
    r = dict(d)
    del r[key]
    return r

# define test model function
def modelfun(xdata, theta):
    m = theta[0]
    b = theta[1]
    nrow = xdata.shape[0]
    y = np.zeros([nrow,1])
    y[:,0] = m*xdata.reshape(nrow,) + b
    return y

def ssfun(theta, data, local = None):
    xdata = data.xdata[0]
    ydata = data.ydata[0]
    # eval model
    ymodel = modelfun(xdata, theta)
    # calc sos
    ss = sum((ymodel[:,0] - ydata[:,0])**2)
    return ss

def basic_mcmc():
    # Initialize MCMC object
    mcstat = MCMC()
    # Add data
    nds = 100
    x = np.linspace(2, 3, num=nds)
    y = 2.*x + 3. + 0.1*np.random.standard_normal(x.shape)
    mcstat.data.add_data_set(x, y)

    mcstat.simulation_options.define_simulation_options(nsimu = int(5.0e3), updatesigma = 1, method = 'dram', verbosity = 0)

    # update model settings
    mcstat.model_settings.define_model_settings(sos_function = ssfun)

    mcstat.parameters.add_model_parameter(name = 'm', theta0 = 2., minimum = -10, maximum = np.inf, sample = 1)
    mcstat.parameters.add_model_parameter(name = 'b', theta0 = -5., minimum = -10, maximum = 100, sample = 0)
    mcstat.parameters.add_model_parameter(name = 'b2', theta0 = -5., minimum = -10, maximum = 100, sample = 1)
    return mcstat

def setup_mcmc():
    mcstat = basic_mcmc()
    mcstat._initialize_simulation()
    
    # extract components
    model = mcstat.model_settings
    options = mcstat.simulation_options
    parameters = mcstat.parameters
    data = mcstat.data
    return model, options, parameters, data

def setup_mcmc_case_mh():
    mcstat = basic_mcmc()
    mcstat._initialize_simulation()
    
    # extract components
    sos_object = mcstat._MCMC__sos_object
    prior_object = mcstat._MCMC__prior_object
    parameters = mcstat.parameters
    return sos_object, prior_object, parameters

def setup_mcmc_case_dr():
    mcstat = basic_mcmc()
    mcstat._initialize_simulation()
    
    # extract components
    model = mcstat.model_settings
    options = mcstat.simulation_options
    parameters = mcstat.parameters
    data = mcstat.data
    covariance = mcstat._covariance
    rejected = {'total': 10, 'outside_bounds': 2}
    chain = np.zeros([options.nsimu, 2])
    s2chain = np.zeros([options.nsimu, 1])
    sschain = np.zeros([options.nsimu, 1])
    return model, options, parameters, data, covariance, rejected, chain, s2chain, sschain

# define test model function
def predmodelfun(data, theta):
    m = theta[0]
    b = theta[1]
    nrow = data.xdata[0].shape[0]
    y = np.zeros([nrow,1])
    y[:,0] = m*data.xdata[0].reshape(nrow,) + b
    return y

def basic_data_structure():
    DS = DataStructure()
    x = np.random.random_sample(size = (100, 1))
    y = np.random.random_sample(size = (100, 1))
    DS.add_data_set(x = x, y = y)
    return DS

def non_basic_data_structure():
    DS = basic_data_structure()
    x = np.random.random_sample(size = (100, 1))
    y = np.random.random_sample(size = (100, 2))
    DS.add_data_set(x = x, y = y)
    return DS

def setup_pseudo_results():
    results = {
            'chain': np.random.random_sample(size = (100,2)),
            's2chain': np.random.random_sample(size = (100,1)),
            'sschain': np.random.random_sample(size = (100,1)),
            'parind': np.array([[0, 1]]),
            'local': np.array([[0, 0]]),
            'model_settings': {'nbatch': np.random.random_sample(size = (100,1))},
            'theta': np.random.random_sample(size = (2,)),
            'sstype': np.random.random_sample(size = (1,1)),
            }
    return results

def setup_pseudo_ci():
    ci = []
    ci1 = []
    ci1.append([np.random.random_sample(size = (100,)),np.random.random_sample(size = (100,)),np.random.random_sample(size = (100,))])
    ci.append(ci1)
    return ci

def setup_mcmc_case_cp(initialize = True):
    mcstat = basic_mcmc()
    if initialize:
        mcstat._initialize_simulation()
    
    return mcstat

def generate_temp_folder():
    tmpfolder = 'temp0'
    count = 0
    flag = True
    while flag is True:
        if os.path.isdir(str('{}'.format(tmpfolder))):
            count += 1
            tmpfolder = str('{}{}'.format('temp',count))
        else:
            flag = False
    return tmpfolder

def generate_temp_file(extension = 'h5'):
    tmpfile = str('temp0.{}'.format(extension))
    count = 0
    flag = True
    while flag is True:
        if os.path.isfile(str('{}'.format(tmpfile))):
            count += 1
            tmpfile = str('{}{}.{}'.format('temp',count,extension))
        else:
            flag = False
    return tmpfile


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 26 10:23:51 2018

@author: prmiles
"""

# import required packages
import numpy as np
import numpy.matlib as npm
import sys
import scipy
import scipy.stats
from scipy.fftpack import fft
from ..plotting.utilities import generate_default_names, extend_names_to_match_nparam

# display chain statistics
def chainstats(chain = None, results = None, returnstats = False):
    '''
    Calculate chain statistics.

    Args:
        * **chain** (:class:`~numpy.ndarray`): Sampling chain.
        * **results** (:py:class:`dict`): Results from MCMC simulation.
        * **returnstats** (:py:class:`bool`): Flag to return statistics.

    Returns:
        * **stats** (:py:class:`dict`): Statistical measures of chain convergence.
    '''
    if chain is None:
        prst = str('No chain reported - run simulation first.')
        print(prst)
        return prst
    else:
        nsimu, npar = chain.shape
        names = get_parameter_names(npar, results)

        # calculate mean and standard deviation of entire chain
        meanii = []
        stdii = []
        for ii in range(npar):
            meanii.append(np.mean(chain[:,ii]))
            stdii.append(np.std(chain[:,ii]))
            
        # calculate batch mean standard deviation
        bmstd = batch_mean_standard_deviation(chain)
        mcerr = bmstd/np.sqrt(nsimu)
            
        # calculate geweke's MCMC convergence diagnostic
        z,p = geweke(chain)
        
        # estimate integrated autocorrelation time
        tau, m = integrated_autocorrelation_time(chain)
        
        # print statistics
        print_chain_statistics(names, meanii, stdii, mcerr, tau, p)
        
        # assign stats to dictionary
        stats = {'mean': list(meanii),
                'std': list(stdii),
                'bmstd': list(bmstd),
                'mcerr': list(mcerr),
                'geweke': {'z': list(z), 'p': list(p)},
                'iact': {'tau': list(tau), 'm': list(m)}
                }
        
        if returnstats is True:
            return stats
          
def print_chain_statistics(names, meanii, stdii, mcerr, tau, p):
    '''
    Print chain statistics to terminal window.

    Args:
        * **names** (:py:class:`list`): List of parameter names.
        * **meanii** (:py:class:`list`): Parameter mean values.
        * **stdii** (:py:class:`list`): Parameter standard deviation.
        * **mcerr** (:class:`~numpy.ndarray`): Normalized batch mean standard deviation.
        * **tau** (:class:`~numpy.ndarray`): Integrated autocorrelation time.
        * **p** (:class:`~numpy.ndarray`): Geweke's convergence diagnostic.

    Example display:

    ::

        ---------------------
        name      :       mean        std     MC_err        tau     geweke
        $p_{0}$   :     1.9680     0.0319     0.0013    36.3279     0.9979
        $p_{1}$   :     3.0818     0.0803     0.0035    37.1669     0.9961
        ---------------------
    '''
    npar = len(names)
    # print statistics
    print('\n---------------------')
    print('{:10s}: {:>10s} {:>10s} {:>10s} {:>10s} {:>10s}'.format('name','mean','std', 'MC_err', 'tau', 'geweke'))
    for ii in range(npar):
        if meanii[ii] > 1e4:
            print('{:10s}: {:10.4g} {:10.4g} {:10.4f} {:10.4f} {:10.4f}'.format(names[ii],meanii[ii],stdii[ii],mcerr[ii],tau[ii],p[ii]))
        else:
            print('{:10s}: {:10.4f} {:10.4f} {:10.4f} {:10.4f} {:10.4f}'.format(names[ii],meanii[ii],stdii[ii],mcerr[ii],tau[ii],p[ii]))
    print('---------------------')
        
# ----------------------------------------------------
def batch_mean_standard_deviation(chain, b = None):
    '''
    Standard deviation calculated from batch means

    Args:
        * **chain** (:class:`~numpy.ndarray`): Sampling chain.
        * **b** (:py:class:`int`): Step size.

    Returns:
        * **s** (:class:`~numpy.ndarray`): Batch mean standard deviation.
    '''
    
    nsimu, npar = chain.shape
    
    # determine step size
    if b is None:
        b = max(10,np.fix(nsimu/20))
        b = int(b)
        
    # define indices
    inds = np.arange(0,nsimu+1,b)
    nb = len(inds)-1
    if nb < 2:
        sys.exit('too few batches')
        
    # initialize batch mean array
    y = np.zeros([nb, npar])
    # calculate mean of each batch
    for ii in range(nb):
        y[ii,:] = np.mean(chain[inds[ii]:inds[ii+1],:],0)
        
    # calculate estimated standard deviation of MC estimate
    s = np.sqrt(sum((y - npm.repmat(np.mean(chain,0),nb,1))**2)/(nb-1)*b)
    
    return s
# ----------------------------------------------------
def geweke(chain, a = 0.1, b = 0.5):
    '''
    Geweke's MCMC convergence diagnostic

    Test for equality of the means of the first a% (default 10%) and
    last b% (50%) of a Markov chain - see :cite:`brooks1998assessing`.

    Args:
        * **chain** (:class:`~numpy.ndarray`): Sampling chain.
        * **a** (:py:class:`float`): First a% of chain.
        * **b** (:py:class:`float`): Last b% of chain.

    Returns:
        * **z** (:class:`~numpy.ndarray`): Convergence diagnostic prior to CDF.
        * **p** (:class:`~numpy.ndarray`): Geweke's MCMC convergence diagnostic.

    .. note::

        The percentage of the chain should be given as a decimal between zero and one.
        So, for the first 10% of the chain, define :code:`a = 0.1`.  Likewise,
        for the last 50% of the chain, define :code:`b = 0.5`.
    '''
    
    nsimu = chain.shape[0]
    
    # determine integer locations based on percentages
    na = int(np.floor(a*nsimu))
    nb = nsimu - int(np.floor(b*nsimu))
    
    # check sizes
    if (na + nb) >= nsimu:
        sys.exit('na + nb = {}, > nsimu = {}'.format(na+nb,nsimu))
    
    # calculate mean chain responses
    m1 = np.mean(chain[0:na,:],0)
    m2 = np.mean(chain[nb:nsimu+1,:],0)
    
    # calculate spectral estimates for variance
    sa = spectral_estimate_for_variance(chain[0:na,:])
    sb = spectral_estimate_for_variance(chain[nb:nsimu+1,:])
    
    z = (m1 - m2)/(np.sqrt(sa/na + sb/(nsimu - nb)))
    p = 2*(1-scipy.stats.norm.cdf(np.abs(z)))
    return z, p
        
        
def spectral_estimate_for_variance(x):
    '''
    Spectral density at frequency zero.

    Args:
        * **x** (:class:`~numpy.ndarray`): Array of points - portion of chain.
    
    Returns:
        * **s** (:class:`~numpy.ndarray`): Spectral estimate for variance.
    '''
    m,n = x.shape
    s = np.zeros([n,])
    for ii in range(n):
        y = power_spectral_density_using_hanning_window(x[:,ii],m)
        s[ii] = y[0]
        
    return s
            
def power_spectral_density_using_hanning_window(x, nfft = None, nw = None):
    '''
    Power spectral density using Hanning window.
    
    Args:
        * **x** (:class:`~numpy.ndarray`): Array of points - portion of chain.
        * **nfft** (:py:class:`int`): Length of Fourier transform.
        * **nw** (:py:class:`int`): Size of window.
    
    Returns:
        * **y** (:class:`~numpy.ndarray`): Power spectral density.
    '''
    if nfft is None:
        nfft = min(len(x),256)
        
    if nw is None:
        nw = int(np.fix(nfft/4))
    
    noverlap = int(np.fix(nw/2))
    
    # define hanning window
    w = 0.5*(1-np.cos(2*np.pi*np.arange(1,nw+1)/(nw+1)))
    
    n = len(x)
    if n < nw:
        x[nw+1] = 0
        n = nw
        
    # number of windows
    k = int(np.fix((n - noverlap)/(nw - noverlap)))
    index = np.arange(nw) # 0,1,...,nw-1
    kmu = k*np.linalg.norm(w)**2 # normalizing scale factor
    
    # initialize array
    y = np.zeros([nfft,1])
    for ii in range(k):
        xw = w*x[index]
        index = index + (nw - noverlap)
        Xx = np.abs(fft(xw,nfft)**2).reshape(nfft,1)
        y = y + Xx
    
    y = y*(1/kmu) # normalize
    n2 = int(np.floor(nfft/2))
    y = y[0:n2]
#    f = 1/n*np.arange(0,n2)
    return y
# ----------------------------------------------------
def integrated_autocorrelation_time(chain):
    '''
    Estimates the integrated autocorrelation time using Sokal's
    adaptive truncated periodogram estimator.

    Args:
        * **chain** (:class:`~numpy.ndarray`): Sampling chain.

    Returns:
        * **tau** (:class:`~numpy.ndarray`): Autocorrelation time.
        * **m** (:class:`~numpy.ndarray`): Counter.
    '''
    # get shape of chain
    npar = chain.shape[1]
    
    # initialize arrays
    tau = np.zeros([npar,])
    m = np.zeros([npar,])
    
    x = fft(chain,axis=0)
    xr = np.real(x)
    xi = np.imag(x)
    xmag = xr**2 + xi**2
    xmag[0,:] = 0.
    xmag = np.real(fft(xmag,axis=0))
    var = xmag[0,:]/len(chain)/(len(chain)-1)
    
    for jj in range(npar):
        if var[jj] == 0:
            continue
        xmag[:,jj] = xmag[:,jj]/xmag[0,jj]
        snum = -1/3
        for ii in range(len(chain)):
            snum = snum + xmag[ii,jj] - 1/6
            if snum < 0:
                tau[jj] = 2*(snum + (ii)/6)
                m[jj] = ii + 1
                break
            
    return tau, m
                
# ----------------------------------------------------
def get_parameter_names(nparam, results):
    '''
    Get parameter names from results dictionary.
    
    If no results found, then default names are generated.  If some results are
    found, then an extended set is generated to complete the list requirement.
    Uses the functions: :func:`~.plotting.utilities.generate_default_names` and
    :func:`~.plotting.utilities.extend_names_to_match_nparam`
    
    Args:
        * **nparam** (:py:class:`int`): Number of parameter names needed
    
    Returns:
        * **names** (:py:class:`list`): List of length `nparam` with strings.
    '''
    if results is None: # results not specified
        names = generate_default_names(nparam)
    else:
        names = results['names']
        names = extend_names_to_match_nparam(names, nparam)
            
    return names


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  1 09:12:06 2018

@author: prmiles
"""
import h5py
import numpy as np
import os
#import fnmatch
import warnings

def print_log_files(savedir):
    '''
    Print log files to screen.

    Args:
        * **savedir** (:py:class:`str`): Directory where log files are saved.
        
    The output display will include a date/time stamp, as well as indices of
    the chain that were saved during that export sequence.
    
    Example display:
    ::
        
        --------------------------
        Display log file: <savedir>/binlogfile.txt
        2018-05-03 14:15:54     0       999
        2018-05-03 14:15:54     1000    1999
        2018-05-03 14:15:55     2000    2999
        2018-05-03 14:15:55     3000    3999
        2018-05-03 14:15:55     4000    4999
        --------------------------
    '''
    # file files in savedir with h5 extension
    bf = []
    for file in os.listdir(savedir):
        if 'log' in file:
            bf.append(file)
            
    nlogf = len(bf) # number of log files found
    
    for ii in range(nlogf):
        logfile = _create_path_without_extension(savedir, bf[ii])
        
        with open(logfile, 'r') as file:
            logstr = file.read()
        
        print('--------------------------')
        print('Display log file: {}'.format(logfile))
        print(logstr)
        print('--------------------------\n')
                
      
def read_in_savedir_files(savedir, extension = 'h5', chainfile = 'chainfile', sschainfile = 'sschainfile', s2chainfile = 's2chainfile', covchainfile = 'covchainfile'):
    '''
    Read in log files from directory.
    
    Args:
        * **savedir** (:py:class:`str`): Directory where log files are saved.
        * **extension** (:py:class:`str`): Extension of files being loaded.
        * **chainfile** (:py:class:`str`): Name of chain log file.
        * **sschainfile** (:py:class:`str`): Name of sschain log file.
        * **s2chainfile** (:py:class:`str`): Name of s2chain log file.
        * **covchainfile** (:py:class:`str`): Name of covchain log file.
    '''
    # file files in savedir with h5 extension
    bf = []
    for file in os.listdir(savedir):
        if file.endswith(str('.{}'.format(extension))):
            bf.append(file)
            
    # create full path names with extension
    chainfile = _create_path_with_extension(savedir, chainfile, extension = extension)
    sschainfile = _create_path_with_extension(savedir, sschainfile, extension = extension)
    s2chainfile = _create_path_with_extension(savedir, s2chainfile, extension = extension)
    covchainfile = _create_path_with_extension(savedir, covchainfile, extension = extension)
   
    if extension == 'h5':
        # read in binary data -> assign to dictionary
        out = {
                'chain': read_in_bin_file(chainfile),
                'sschain': read_in_bin_file(sschainfile),
                's2chain': read_in_bin_file(s2chainfile),
                'covchain': read_in_bin_file(covchainfile)
                }
    elif extension == 'txt':
        # read in text data -> assign to dictionary
        out = {
                'chain': read_in_txt_file(chainfile),
                'sschain': read_in_txt_file(sschainfile),
                's2chain': read_in_txt_file(s2chainfile),
                'covchain': read_in_txt_file(covchainfile)
                }
    else:
        out = None
        warnings.warn('Unknown extension specified -> log files saved as either h5 (binary) or txt (text).')
    
    return out
    
def read_in_parallel_savedir_files(parallel_dir, extension = 'h5', chainfile = 'chainfile', sschainfile = 'sschainfile', s2chainfile = 's2chainfile', covchainfile = 'covchainfile'):
    '''
    Read in log files from directory containing results from parallel MCMC simulation.
    
    Args:
        * **parallel_dir** (:py:class:`str`): Directory where parallel log files are saved.
        * **extension** (:py:class:`str`): Extension of files being loaded.
        * **chainfile** (:py:class:`str`): Name of chain log file.
        * **sschainfile** (:py:class:`str`): Name of sschain log file.
        * **s2chainfile** (:py:class:`str`): Name of s2chain log file.
        * **covchainfile** (:py:class:`str`): Name of covchain log file.
    '''
    # find folders in parallel_dir with name 'chain_#'
    chainfolders = os.listdir(parallel_dir)
    out = []
    for folder in chainfolders:
        # create full path names with extension
        savedir = _create_path_without_extension(parallel_dir, folder)
        out.append(read_in_savedir_files(savedir, extension = extension, chainfile = chainfile, sschainfile = sschainfile, s2chainfile = s2chainfile, covchainfile = covchainfile))
        
    return out
    
def read_in_bin_file(filename):
    '''
    Read in information from file containing binary data.
    
    If file exists, it will read in the array elements.  Otherwise, it will return
    and empty list.
    
    Args:
        * **filename** (:py:class:`str`): Name of file to read.

    Returns:
        * **out** (:class:`~numpy.ndarray`): Array of chain elements.
    '''
    try:
        hf = h5py.File(filename, 'r')
        ds = list(hf.keys()) # data sets
        # check size of each set
        sh = np.zeros([len(ds),2], dtype=int)
        for ii, dsii in enumerate(ds):
            tmp = hf.get(ds[ii])
            sh[ii,:] = tmp.shape
            
        # initialize array for merged datasets
        # assume all have same number of columns
        out = np.zeros([sum(sh[:,0]),sh[0,1]])
        for ii, dsii in enumerate(ds):
            if ii == 0:
                a = 0;
                b = sh[ii,0]
            else:
                a = a + sh[ii-1,0]
                b = b + sh[ii-1,0]
                
            out[a:b,:] = np.array(hf.get(dsii))
            
        hf.close()
    except Exception as ex:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)
        print(message)
        warnings.warn('Exception raised reading {} - does not exist - check path/file name.  Assigning to empty list.'.format(filename))
        out = []
    
    return out
    
def read_in_txt_file(filename):
    '''
    Read in information from file containing text data.
    
    If file exists, it will read in the array elements.  Otherwise, it will return
    and empty list.
    
    Args:
        * **filename** (:py:class:`str`): Name of file to read.
    
    Returns:
        * **out** (:class:`~numpy.ndarray`): Array of chain elements.
    '''
    try:
        out = np.loadtxt(filename)
    except Exception as ex:
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)
        print(message)
        warnings.warn('Exception raised reading {} - does not exist - check path/file name.  Assigning to empty list.'.format(filename))
        out = []
    
    return out
    
def _create_path_with_extension_for_all_logs(options, extension = 'h5'):
    chainfile = _create_path_with_extension(options.savedir, options.chainfile, extension)
    s2chainfile = _create_path_with_extension(options.savedir, options.s2chainfile, extension)
    sschainfile = _create_path_with_extension(options.savedir, options.sschainfile, extension)
    covchainfile = _create_path_with_extension(options.savedir, options.covchainfile, extension)
    return chainfile, s2chainfile, sschainfile, covchainfile
    
def _create_path_with_extension(savedir, file, extension = 'h5'):
    file = os.path.join(savedir, str('{}.{}'.format(file,extension)))
    return file
    
def _create_path_without_extension(savedir, file):
    file = os.path.join(savedir, file)
    return file
    
def _add_to_log(filename, logstr):
    with open(filename, 'a') as logfile:
        logfile.write(logstr)
    
def _save_to_bin_file(filename, datasetname, mtx):
    hf = h5py.File(filename, 'a')
    hf.create_dataset(datasetname, data = mtx)
    hf.close()
        
def _save_to_txt_file(filename, mtx):
    handle = open(filename, 'ab')
    np.savetxt(handle,mtx)
    handle.close()
    
def _check_directory(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 11 13:17:23 2018

Description: This file contains a series of utilities designed to test the
features in the 'ModelSettings.py" package of the pymcmcstat module.

@author: prmiles
"""

from pymcmcstat.settings.DataStructure import DataStructure
from pymcmcstat.settings.SimulationOptions import SimulationOptions
from pymcmcstat.settings.ModelSettings import ModelSettings
import unittest
import numpy as np

MS = ModelSettings()

def setup_model_settings(**kwargs):
    # create settings
    MS = ModelSettings()
    MS.define_model_settings(**kwargs)
    return MS

def setup_simulation_options(**kwargs):
    # create options
    options = SimulationOptions()
    options.define_simulation_options(**kwargs)
    return options

# --------------------------
# initialization
# --------------------------
class Model_Settings_Initialization(unittest.TestCase):

    def test_does_initialization_yield_description(self):
        ms = ModelSettings()
        self.assertTrue(hasattr(ms,'description'))

# --------------------------
# define_model_settings
# --------------------------
class Define_Model_Settings(unittest.TestCase):
       
    def test_default_sos_function(self):
        ms = setup_model_settings()
        self.assertEqual(ms.sos_function, None, msg = 'Should be ''None''')
        
    def test_default_prior_function(self):
        ms = setup_model_settings()
        self.assertEqual(ms.prior_function, None, msg = 'Should be ''None''')
        
    def test_default_prior_type(self):
        ms = setup_model_settings()
        self.assertEqual(ms.prior_type, 1, msg = 'Should be 1')
        
    def test_default_prior_update_function(self):
        ms = setup_model_settings()
        self.assertEqual(ms.prior_update_function, None, msg = 'Should be ''None''')
        
    def test_default_prior_pars(self):
        ms = setup_model_settings()
        self.assertEqual(ms.prior_pars, None, msg = 'Should be ''None''')
        
    def test_default_model_function(self):
        ms = setup_model_settings()
        self.assertEqual(ms.model_function, None, msg = 'Should be ''None''')
        
    def test_default_sigma2(self):
        ms = setup_model_settings()
        self.assertEqual(ms.sigma2, None, msg = 'Should be ''None''')
        
    def test_default_N(self):
        ms = setup_model_settings()
        self.assertEqual(ms.N, None, msg = 'Should be ''None''')
        
    def test_default_S20(self):
        ms = setup_model_settings()
        self.assertTrue(np.isnan(ms.S20), msg = 'Should be ''nan''')
        
    def test_default_N0(self):
        ms = setup_model_settings()
        self.assertEqual(ms.N0, None, msg = 'Should be ''None''')
        
    def test_default_nbatch(self):
        ms = setup_model_settings()
        self.assertEqual(ms.nbatch, None, msg = 'Should be ''None''')
        
    def test_list_N_returned_as_numpy_array(self):
        ms = setup_model_settings(N = [1, 1])
        self.assertIsInstance(ms.N, np.ndarray, msg = 'Should be a numpy array')
        
    def test_float_N_returned_as_numpy_array(self):
        ms = setup_model_settings(N = 1.2)
        self.assertIsInstance(ms.N, np.ndarray, msg = 'Should be a numpy array')
        
    def test_int_N_returned_as_numpy_array(self):
        ms = setup_model_settings(N = 1)
        self.assertIsInstance(ms.N, np.ndarray, msg = 'Should be a numpy array')
        
    def test_tuple_N_returns_error(self):
        ms = ModelSettings()
        with self.assertRaises(SystemExit, msg = 'Tuple input not accepted'):
            ms.define_model_settings(N = (1,2))
            
# --------------------------
# _check_dependent_model_settings
# --------------------------
class Check_Dependent_Model_Settings(unittest.TestCase):
    
    def test_nbatch_calc_from_data(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.nbatch, 1, msg = 'nbatch should equal number of calls to ''add_data_set''')
        
    def test_nbatch_calc_from_data_y_2d(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.nbatch, 1, msg = 'nbatch should equal number of calls to ''add_data_set''')
        
    def test_nbatch_calc_from_data_multiple_sets(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.nbatch, 2, msg = 'nbatch should equal number of calls to ''add_data_set''')
        
    def test_nbatch_calc_from_data_multiple_data_sets_of_different_size(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        y = np.zeros([2,])
        data.add_data_set(x,y)
        y = np.zeros([5])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.nbatch, 3, msg = 'nbatch should equal number of calls to ''add_data_set''')
        
    def test_N_calc_from_data(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertTrue(np.array_equal(ms.N, np.array([2])), msg = 'N should equal total number of rows in each y set')
        
    def test_N_calc_from_data_with_multiple_sets(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        y = np.zeros([3,3])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertTrue(np.array_equal(ms.N, np.array([2,3])), msg = 'N should equal total number of rows in each y set')
        
    def test_N_calc_from_data_with_multiple_sets_of_different_size(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        y = np.zeros([3,3])
        data.add_data_set(x,y)
        y = np.zeros([4,7])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertTrue(np.array_equal(ms.N, np.array([2,3,4])), msg = 'N should equal total number of rows in each y set')
        
    def test_N_conflict_produces_error_message(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N = 1)
                
        # calculate dependencies
        with self.assertRaises(SystemExit, msg = 'Conflicting N'):
            ms._check_dependent_model_settings(data, options)
            
    def test_default_sigma2_dependent(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.sigma2, np.ones([1]), msg='default is 1');
        
    def test_default_N0_dependent(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.N0, np.zeros([1]), msg='default is 0');
        
    def test_default_N0_dependent_for_non_default_sigma2(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(sigma2 = 1)
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.N0, np.ones([1]), msg='default is 1');
        
    def test_default_sigma2_dependent_for_non_default_N0(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = 1)
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.sigma2, np.ones(ms.nbatch), msg='default is ones(ms.nbatch)');
        
    def test_default_sigma2_dependent_for_non_default_N0_S20(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = 1, S20 = 1)
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertEqual(ms.sigma2, ms.S20, msg='default is S20');
        
    def test_default_sigma2_dependent_for_non_default_N0_S20_multiple_data_sets(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        y = np.zeros([3,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = 1, S20 = 1)
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        comp = np.linalg.norm(ms.sigma2 - ms.S20)
        self.assertAlmostEqual(comp, 0, msg='default is S20')
        
    def test_default_S20_dependent(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        self.assertAlmostEqual(ms.S20, ms.sigma2, msg='default is sigma2')
        
    def test_default_S20_dependent_for_multiple_data_sets(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2,2])
        data.add_data_set(x,y)
        y = np.zeros([3,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        comp = np.linalg.norm(ms.sigma2 - ms.S20)
        self.assertAlmostEqual(comp, 0, msg='default is sigma2')
        
# --------------------------
# _check_dependent_model_settings_wrt_nsos
# --------------------------
class Check_Dependent_Model_Settings_WRT_Nsos(unittest.TestCase):
    
    def test_size_S20_wrt_nsos(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.S20), nsos, msg = 'length of S20 should equal number of elements returned from sos function')
        
    def test_size_N0_wrt_nsos(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.N0), nsos, msg = 'length of N0 should equal number of elements returned from sos function')
        
    def test_size_sigma2_wrt_nsos(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.sigma2), nsos, msg = 'length of sigma2 should equal number of elements returned from sos function')
    
    def test_size_S20_wrt_nsos_for_non_default_S20(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(S20 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.S20), nsos, msg = 'length of S20 should equal number of elements returned from sos function')
        
    def test_size_S20_wrt_nsos_for_non_default_S20_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(S20 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 3
        with self.assertRaises(SystemExit, msg = 'S20 size mismatch'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
            
    def test_size_N0_wrt_nsos_for_non_default_N0(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.N0), nsos, msg = 'length of N0 should equal number of elements returned from sos function')
    
    def test_size_N0_wrt_nsos_for_non_default_N0_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 3
        with self.assertRaises(SystemExit, msg = 'N0 size mismatch'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
    def test_size_sigma2_wrt_nsos_for_non_default_sigma2(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(sigma2 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.sigma2), nsos, msg = 'length of sigma2 should equal number of elements returned from sos function')
        
    def test_size_sigma2_wrt_nsos_for_non_default_sigma2_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(sigma2 = [1,2])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 3
        with self.assertRaises(SystemExit, msg = 'sigma2 size mismatch'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
            
    def test_value_S20_wrt_nsos_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        # create options
        options = SimulationOptions()
        options.define_simulation_options(nsimu = int(1000))
        
        # create settings
        ms = setup_model_settings(S20 = [1,2])
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        nsos = 1
        with self.assertRaises(SystemExit, msg = 'S20 should not be larger than nsos'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
            
    def test_value_N0_wrt_nsos_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N0 = [1,2])
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        nsos = 1
        with self.assertRaises(SystemExit, msg = 'N0 should not be larger than nsos'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
            
    def test_value_sigma2_wrt_nsos_raises_error(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(sigma2 = [1,2])
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        nsos = 1
        with self.assertRaises(SystemExit, msg = 'sigma2 should not be larger than nsos'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
     
#   CHECK CALCULATION OF NUMBER OF OBSERVATIONS
    def test_size_N_wrt_nsos(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.N), nsos, msg = 'length of N should equal number of elements returned from sos function')
       
    def test_size_N_wrt_nsos_for_2_data_sets(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings()
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 2
        print('N = {}, nsos = {}'.format(ms.N, nsos))
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        print('N = {}, nsos = {}'.format(ms.N, nsos))
        self.assertEqual(len(ms.N), nsos, msg = 'length of N should equal number of elements returned from sos function')
       
    def test_size_N_wrt_nsos_for_non_default_N(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        y = np.zeros([4,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N = [2,4])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 3
        
        with self.assertRaises(SystemExit, msg = 'length of N should equal number of elements returned from sos function'):
            ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
            
    def test_size_N_wrt_nsos_for_non_default_N_and_nsos_equal_1(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        y = np.zeros([4,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N = [2,4])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 1
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(len(ms.N), nsos, msg = 'length of N should equal number of elements returned from sos function')
    
    def test_value_N_wrt_nsos_for_non_default_N_and_nsos_equal_1(self):
        # create test data
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        y = np.zeros([4,2])
        data.add_data_set(x,y)
        
        options = setup_simulation_options(nsimu = int(1000))
        ms = setup_model_settings(N = [2,4])
        
        # calculate dependencies
        ms._check_dependent_model_settings(data, options)
        
        nsos = 1
        ms._check_dependent_model_settings_wrt_nsos(nsos = nsos)
        
        self.assertEqual(ms.N, 6, msg = 'length of N should equal number of elements returned from sos function')

# -------------------------------------------
class ArrayType(unittest.TestCase):
    
    def test_none(self):
        self.assertTrue(MS._array_type(x = None) is None)
    
    def test_int(self):
        self.assertTrue(np.array_equal(MS._array_type(x = 3), np.array([3])))
        
    def test_float(self):
        self.assertTrue(np.array_equal(MS._array_type(x = 3.), np.array([3.])))
        
    def test_list(self):
        self.assertTrue(np.array_equal(MS._array_type(x = [3., 2.]), np.array([3., 2.])))
    
    def test_nparray(self):
        self.assertTrue(np.array_equal(MS._array_type(x = np.array([3., 2.])), np.array([3., 2.])))
        
    def test_dict(self):
        with self.assertRaises(SystemExit, msg = 'Dictionary not expected.'):
            MS._array_type(x = {'hello': 'world'})
            
# -------------------------------------------
class NumberOfObservations(unittest.TestCase):
    
    def test_array_equal(self):
        ms = ModelSettings()
        
        udN = np.ones([1,2])
        dsN = np.ones([1,2])
        
        N = ms._check_number_of_observations(udN = udN, dsN = dsN)
        
        self.assertTrue((N == udN).all() and (N == dsN).all())
        
    def test_len_dsN_gt_udN_case_1(self):
        ms = ModelSettings()
        udN = np.ones([1])
        dsN = np.ones([1,2])
        N = ms._check_number_of_observations(udN = udN, dsN = dsN)
        
        self.assertTrue((N == dsN).all())
        
    def test_len_dsN_gt_udN_case_2(self):
        ms = ModelSettings()
        udN = np.ones([1,2])
        dsN = np.ones([1,3])
        with self.assertRaises(SystemExit, msg = 'Mismatching dimensions'):
            ms._check_number_of_observations(udN = udN, dsN = dsN)
            
    def test_len_udN_gt_dsN_case_1(self):
        ms = ModelSettings()
        udN = np.ones([1,2])
        dsN = np.ones([1])
        N = ms._check_number_of_observations(udN = udN, dsN = dsN)
        
        self.assertTrue((N == dsN).all())
        
    def test_len_udN_gt_dsN_case_2(self):
        ms = ModelSettings()
        udN = np.ones([1,3])
        dsN = np.ones([1,2])
        with self.assertRaises(SystemExit, msg = 'Mismatching dimensions'):
            ms._check_number_of_observations(udN = udN, dsN = dsN)
            
    def test_len_udN_dsN_gt_1(self):
        ms = ModelSettings()
        udN = np.ones([1,3])
        dsN = np.zeros([1,3])
        with self.assertRaises(SystemExit, msg = 'Dimensions match, but values differ.'):
            ms._check_number_of_observations(udN = udN, dsN = dsN)
        
# -------------------------------------------
class DisplayModelSettings(unittest.TestCase):
    
    def test_print_these_none(self):
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = SimulationOptions()
        
        ms = ModelSettings()
        ms.define_model_settings()
        ms._check_dependent_model_settings(data, options)
        ms._check_dependent_model_settings_wrt_nsos(nsos = 1)
        
        print_these = ms.display_model_settings(print_these = None)
        self.assertEqual(print_these, ['sos_function', 'model_function', 'sigma2', 'N', 'N0', 'S20', 'nsos', 'nbatch'], msg = 'Default print keys')
        
    def test_print_these_not_none(self):
        data = DataStructure()
        x = np.zeros([2])
        y = np.zeros([2])
        data.add_data_set(x,y)
        
        options = SimulationOptions()
        
        ms = ModelSettings()
        ms.define_model_settings()
        ms._check_dependent_model_settings(data, options)
        ms._check_dependent_model_settings_wrt_nsos(nsos = 1)
        
        print_these = ms.display_model_settings(print_these = ['model_function'])
        self.assertEqual(print_these, ['model_function'], msg = 'Specified print keys')
# -------------------------------------------


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun 21 10:10:43 2018

@author: prmiles
"""

from pymcmcstat.chain import ChainProcessing as CP
import test.general_functions as gf
import unittest
from mock import patch
import numpy as np
import os
import shutil
import io
import sys

def setup_case():
    mcstat = gf.basic_mcmc()
    mcstat._MCMC__chain = np.random.random_sample(size = (100,2))
    mcstat._MCMC__sschain = np.random.random_sample(size = (100,2))
    mcstat._MCMC__s2chain = np.random.random_sample(size = (100,2))
    mcstat._covariance._R = np.array([[0.5, 0.2],[0., 0.3]])
    return mcstat

# --------------------------
class CreatePathWithExtensionforAllLogs(unittest.TestCase):
    
    def test_cpwe_creation_with_h5(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.simulation_options.savedir = 'test'
        chainfile, s2chainfile, sschainfile, covchainfile = CP._create_path_with_extension_for_all_logs(options = mcstat.simulation_options, extension = 'h5')
        self.assertEqual(chainfile, 'test/chainfile.h5', msg = 'Expect matching string')
        self.assertEqual(s2chainfile, 'test/s2chainfile.h5', msg = 'Expect matching string')
        self.assertEqual(sschainfile, 'test/sschainfile.h5', msg = 'Expect matching string')
        self.assertEqual(covchainfile, 'test/covchainfile.h5', msg = 'Expect matching string')
        
    def test_cpwe_creation_with_txt(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.simulation_options.savedir = 'test'
        chainfile, s2chainfile, sschainfile, covchainfile = CP._create_path_with_extension_for_all_logs(options = mcstat.simulation_options, extension = 'txt')
        self.assertEqual(chainfile, 'test/chainfile.txt', msg = 'Expect matching string')
        self.assertEqual(s2chainfile, 'test/s2chainfile.txt', msg = 'Expect matching string')
        self.assertEqual(sschainfile, 'test/sschainfile.txt', msg = 'Expect matching string')
        self.assertEqual(covchainfile, 'test/covchainfile.txt', msg = 'Expect matching string')
        
# --------------------------
class CreatePathWithExtension(unittest.TestCase):
    def test_cpwe_basic_h5(self):
        file = CP._create_path_with_extension(savedir = 'test', file = 'file', extension = 'h5')
        self.assertEqual(file, 'test/file.h5', msg = 'Expect matching string')
        
    def test_cpwe_basic_txt(self):
        file = CP._create_path_with_extension(savedir = 'test', file = 'file', extension = 'txt')
        self.assertEqual(file, 'test/file.txt', msg = 'Expect matching string')
        
# --------------------------
class CreatePathWithOutExtension(unittest.TestCase):
    def test_cpwe_basic(self):
        file = CP._create_path_without_extension(savedir = 'test', file = 'file')
        self.assertEqual(file, 'test/file', msg = 'Expect matching string')

# --------------------------
class ReadInBinFile(unittest.TestCase):
    def test_readinbinaryfile_warning(self):
        out = CP.read_in_bin_file(filename = 'abcdedf0123')
        self.assertEqual(out, [], msg = 'expect empty list')
    
    def test_readingbinaryfile_array(self):
        tmpfile = gf.generate_temp_file(extension = 'h5')
        CP._save_to_bin_file(filename = tmpfile, datasetname = 'd1', mtx = np.array([[0.1, 0.2]]))
        self.assertTrue(os.path.isfile(tmpfile), msg = 'File exists')
        out = CP.read_in_bin_file(filename = tmpfile)
        self.assertTrue(np.array_equal(out, np.array([[0.1, 0.2]])), msg = str('Expected array match: {}'.format(out)))
        os.remove(tmpfile)
        
# --------------------------
class ReadInTextFile(unittest.TestCase):
    def test_readintextfile_warning(self):
        out = CP.read_in_txt_file(filename = 'abcdedf0123')
        self.assertEqual(out, [], msg = 'expect empty list')
    
    @patch('numpy.loadtxt', return_value = np.array([0.1, 0.2]))
    def test_readingtextfile_array(self, mock_load):
        out = CP.read_in_txt_file(filename = 'abcdedf0123')
        self.assertTrue(np.array_equal(out, np.array([0.1, 0.2])), msg = str('Expected array match: {}'.format(out)))

# -------------------
class CheckDirectory(unittest.TestCase):
    def test_check_directory(self):
        tmpfolder = gf.generate_temp_folder()
        CP._check_directory(tmpfolder)
        self.assertTrue(os.path.isdir(tmpfolder), msg = 'Directory exists')
        os.removedirs(tmpfolder)
        
# -------------------
class CheckSaveToTextFile(unittest.TestCase):
    def test_save_to_text_file(self):
        tmpfile = gf.generate_temp_file(extension = 'txt')
        CP._save_to_txt_file(filename = tmpfile, mtx = np.array([0.1, 0.2]))
        self.assertTrue(os.path.isfile(tmpfile), msg = 'File exists')
        loadmtx = np.loadtxt(tmpfile)
        self.assertTrue(np.array_equal(loadmtx, np.array([0.1, 0.2])), msg = str('Arrays should match: {}'.format(loadmtx)))
        os.remove(tmpfile)
        
# -------------------
class CheckSaveToBinaryFile(unittest.TestCase):
    def test_save_to_binary_file(self):
        tmpfile = gf.generate_temp_file(extension = 'h5')
        CP._save_to_bin_file(filename = tmpfile, datasetname = 'd1', mtx = np.array([0.1, 0.2]))
        self.assertTrue(os.path.isfile(tmpfile), msg = 'File exists')
        os.remove(tmpfile)
        
# -------------------
class AddToLog(unittest.TestCase):
    def test_add_to_log(self):
        tmpfile = gf.generate_temp_file(extension = 'txt')
        CP._add_to_log(filename = tmpfile, logstr = 'hello world')
        self.assertTrue(os.path.isfile(tmpfile), msg = 'File exists')
        with open(tmpfile, 'r') as file:
            loadstr = file.read()
        self.assertEqual(loadstr, 'hello world', msg = 'Message should match')
        os.remove(tmpfile)
# -------------------
class ReadInSavedirFiles(unittest.TestCase):
    
    def test_read_in_savedir_files_h5(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        mcstat._MCMC__save_chains_to_bin(start = 0, end = 100)
        
        out = CP.read_in_savedir_files(savedir, extension = 'h5')
        self.assertTrue(np.array_equal(out['chain'], mcstat._MCMC__chain), msg = str('Expect arrays to match: chain'))
        self.assertTrue(np.array_equal(out['sschain'], mcstat._MCMC__sschain), msg = str('Expect arrays to match: sschain'))
        self.assertTrue(np.array_equal(out['s2chain'], mcstat._MCMC__s2chain), msg = str('Expect arrays to match: s2chain'))
        self.assertTrue(np.array_equal(out['covchain'], np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R)), msg = str('Expect arrays to match: chain'))
        shutil.rmtree(savedir)
        
    def test_read_in_savedir_files_txt(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        
        out = CP.read_in_savedir_files(savedir, extension = 'txt')
        self.assertTrue(np.array_equal(out['chain'], mcstat._MCMC__chain), msg = str('Expect arrays to match: chain'))
        self.assertTrue(np.array_equal(out['sschain'], mcstat._MCMC__sschain), msg = str('Expect arrays to match: sschain'))
        self.assertTrue(np.array_equal(out['s2chain'], mcstat._MCMC__s2chain), msg = str('Expect arrays to match: s2chain'))
        self.assertTrue(np.array_equal(out['covchain'], np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R)), msg = str('Expect arrays to match: chain'))
        shutil.rmtree(savedir)
        
    def test_read_in_savedir_files_unknown(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        
        out = CP.read_in_savedir_files(savedir, extension = 'unknown')
        self.assertEqual(out, None, msg = 'Expect None')
        shutil.rmtree(savedir)

# -------------------
class ReadInParallelSavedirFiles(unittest.TestCase):
    def chain_comp(self, out, mcstat):
        self.assertTrue(np.array_equal(out['chain'], mcstat._MCMC__chain), msg = str('Expect arrays to match: chain'))
        self.assertTrue(np.array_equal(out['sschain'], mcstat._MCMC__sschain), msg = str('Expect arrays to match: sschain'))
        self.assertTrue(np.array_equal(out['s2chain'], mcstat._MCMC__s2chain), msg = str('Expect arrays to match: s2chain'))
        self.assertTrue(np.array_equal(out['covchain'], np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R)), msg = str('Expect arrays to match: chain'))

    def test_read_in_parallel_bin(self):
        mcstat = setup_case()
        parallel_dir = gf.generate_temp_folder()
        for ii in range(3):
            chain_dir = str('chain_{}'.format(ii))
            mcstat.simulation_options.savedir = str('{}{}{}'.format(parallel_dir,os.sep,chain_dir))
            mcstat._MCMC__save_chains_to_bin(start = 0, end = 100)
        
        out = CP.read_in_parallel_savedir_files(parallel_dir = parallel_dir, extension = 'h5')
        self.chain_comp(out[0], mcstat)
        self.chain_comp(out[1], mcstat)
        self.chain_comp(out[2], mcstat)
        shutil.rmtree(parallel_dir)
        
    def test_read_in_parallel_txt(self):
        mcstat = setup_case()
        parallel_dir = gf.generate_temp_folder()
        for ii in range(3):
            chain_dir = str('chain_{}'.format(ii))
            mcstat.simulation_options.savedir = str('{}{}{}'.format(parallel_dir,os.sep,chain_dir))
            mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        
        out = CP.read_in_parallel_savedir_files(parallel_dir = parallel_dir, extension = 'txt')
        self.chain_comp(out[0], mcstat)
        self.chain_comp(out[1], mcstat)
        self.chain_comp(out[2], mcstat)
        shutil.rmtree(parallel_dir)
        
    def test_read_in_parallel_unknown(self):
        mcstat = setup_case()
        parallel_dir = gf.generate_temp_folder()
        for ii in range(3):
            chain_dir = str('chain_{}'.format(ii))
            mcstat.simulation_options.savedir = str('{}{}{}'.format(parallel_dir,os.sep,chain_dir))
            mcstat._MCMC__save_chains_to_bin(start = 0, end = 100)
        
        out = CP.read_in_parallel_savedir_files(parallel_dir = parallel_dir, extension = 'unknown')
        self.assertEqual(out[0], None)
        self.assertEqual(out[1], None)
        self.assertEqual(out[2], None)
        shutil.rmtree(parallel_dir)
# -------------------
class PrintLogFiles(unittest.TestCase):
    def test_print_log_files(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        
        capturedOutput = io.StringIO()
        sys.stdout = capturedOutput
        CP.print_log_files(savedir)
        sys.stdout = sys.__stdout__
        
        self.assertTrue(isinstance(capturedOutput.getvalue(), str), msg = 'Should contain a string')
        self.assertTrue('Display log file:' in capturedOutput.getvalue(), msg = 'Expect string to contain these works')
        shutil.rmtree(savedir)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May 31 05:12:38 2018

@author: prmiles
"""

from pymcmcstat.procedures.SumOfSquares import SumOfSquares
import test.general_functions as gf
import unittest
import numpy as np

# --------------------------
class InitializeSOS(unittest.TestCase):

    def test_init_sos(self):
        model, __, parameters, data = gf.setup_mcmc()
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        SOSD = SOS.__dict__
        check = {'nbatch': 1, 'parind': parameters._parind, 'value': parameters._value,
                 'local': parameters._local, 'data': data, 'model_function': model.model_function,
                 'sos_function': model.sos_function, 'sos_style': 1}
        items = ['nbatch', 'sos_style', 'sos_function', 'model_function', 'data']
        for ii, ai in enumerate(items):
            self.assertEqual(SOSD[ai], check[ai], msg = str('{}: {} != {}'.format(ai, SOSD[ai], check[ai])))
            
        array_items = ['parind', 'value', 'local']
        for ii, ai in enumerate(array_items):
            self.assertTrue(np.array_equal(SOSD[ai], check[ai]), msg = str('{}: {} != {}'.format(ai, SOSD[ai], check[ai])))
            
    def test_init_sos_function_is_none(self):
        model, options, parameters, data = gf.setup_mcmc()
        model.sos_function = None
        model.model_function = gf.modelfun
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        SOSD = SOS.__dict__
        check = {'nbatch': 1, 'parind': parameters._parind, 'value': parameters._value,
                 'local': parameters._local, 'data': data, 'model_function': gf.modelfun,
                 'sos_function': None, 'sos_style': 4}
        items = ['nbatch', 'sos_style', 'sos_function', 'model_function', 'data']
        for ii, ai in enumerate(items):
            self.assertEqual(SOSD[ai], check[ai], msg = str('{}: {} != {}'.format(ai, SOSD[ai], check[ai])))
            
        array_items = ['parind', 'value', 'local']
        for ii, ai in enumerate(array_items):
            self.assertTrue(np.array_equal(SOSD[ai], check[ai]), msg = str('{}: {} != {}'.format(ai, SOSD[ai], check[ai])))
        
    def test_init_sos_model_function_is_none(self):
        model, options, parameters, data = gf.setup_mcmc()
        model.sos_function = None
        model.model_function = None
        with self.assertRaises(SystemExit):
            SumOfSquares(model = model, data = data, parameters = parameters)
            
class EvaluateSOS(unittest.TestCase):

    def test_eval_sos(self):
        model, options, parameters, data = gf.setup_mcmc()
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        theta = np.array([2., 5.])
        ss = SOS.evaluate_sos_function(theta = theta)
        self.assertTrue(isinstance(ss, np.ndarray), msg = 'Expect numpy array return')
        self.assertEqual(ss.size, 1, msg = 'Size of array is 1')
        self.assertTrue(isinstance(ss[0], float), msg = 'Numerical result returned')
        self.assertTrue(np.array_equal(SOS.value[SOS.parind], theta), msg = 'Value(s) updated')
        
    def test_eval_sos_none(self):
        model, options, parameters, data = gf.setup_mcmc()
        model.sos_function = None
        model.model_function = gf.modelfun
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        theta = np.array([2., 5.])
        ss = SOS.evaluate_sos_function(theta = theta)
        self.assertTrue(isinstance(ss, np.ndarray), msg = 'Expect numpy array return')
        self.assertEqual(ss.size, 1, msg = 'Size of array is 1')
        self.assertTrue(isinstance(ss[0], float), msg = 'Numerical result returned')
        self.assertTrue(np.array_equal(SOS.value[SOS.parind], theta), msg = 'Value(s) updated')
        
    def test_eval_sos_model_none(self):
        model, options, parameters, data = gf.setup_mcmc()
        SOS = SumOfSquares(model = model, data = data, parameters = parameters)
        SOS.sos_style = 2
        theta = np.array([2., 5.])
        ss = SOS.evaluate_sos_function(theta = theta)
        self.assertTrue(isinstance(ss, np.ndarray), msg = 'Expect numpy array return')
        self.assertEqual(ss.size, 1, msg = 'Size of array is 1')
        self.assertTrue(isinstance(ss[0], float), msg = 'Numerical result returned')
        self.assertTrue(np.array_equal(SOS.value[SOS.parind], theta), msg = 'Value(s) updated')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  6 08:33:47 2018

@author: prmiles
"""

from pymcmcstat.MCMC import print_rejection_statistics, MCMC
from pymcmcstat.structures.ParameterSet import ParameterSet
from pymcmcstat.chain import ChainProcessing
import test.general_functions as gf
import unittest
from mock import patch
import io
import sys
import os
import shutil
import numpy as np

def setup_pseudo_results(initialize = True):
    mcstat = gf.setup_mcmc_case_cp(initialize = initialize)
    rejectedin = {'total': 10, 'in_adaptation_interval': 4, 'outside_bounds': 1}
    mcstat._MCMC__rejected = rejectedin.copy()
    mcstat._MCMC__simulation_time = 0.1
    return mcstat

def setup_case():
    mcstat = gf.basic_mcmc()
    mcstat._MCMC__chain = np.random.random_sample(size = (100,2))
    mcstat._MCMC__sschain = np.random.random_sample(size = (100,2))
    mcstat._MCMC__s2chain = np.random.random_sample(size = (100,2))
    mcstat._covariance._R = np.array([[0.5, 0.2],[0., 0.3]])
    return mcstat

# --------------------------
class MCMCInitialization(unittest.TestCase):
    
    def test_initialization(self):
        MC = MCMC()
        
        check_these = ['data', 'model_settings', 'simulation_options', 'parameters',
                       '_error_variance', '_covariance', '_sampling_methods', '_mcmc_status']
        for ct in check_these:
            self.assertTrue(hasattr(MC, ct), msg = str('Object missing attribute: {}'.format(ct)))
        
        self.assertFalse(MC._mcmc_status, msg = 'Status is False')

# --------------------------
class DisplayCurrentMCMCSettings(unittest.TestCase):
    
    def test_standard_print(self):
        mcstat = gf.setup_mcmc_case_cp()
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        mcstat.display_current_mcmc_settings()
        sys.stdout = sys.__stdout__                     # Reset redirect.
        self.assertTrue(isinstance(capturedOutput.getvalue(), str), msg = 'Caputured string')
        
# --------------------------
class PrintRejectionStatistics(unittest.TestCase):
    def test_print_stats(self):
        rejected = {'total': 10, 'in_adaptation_interval': 4, 'outside_bounds': 1}
        isimu = 100
        iiadapt = 10
        verbosity = 3
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        print_rejection_statistics(rejected = rejected, isimu = isimu, iiadapt = iiadapt, verbosity = verbosity)
        sys.stdout = sys.__stdout__
        
        self.assertEqual(capturedOutput.getvalue(), str('i:{} ({},{},{})\n\n'.format(
                isimu, rejected['total']*isimu**(-1)*100, rejected['in_adaptation_interval']*iiadapt**(-1)*100,
                rejected['outside_bounds']*isimu**(-1)*100)), msg = 'Strings should match')
        
# --------------------------
class UpdateChain(unittest.TestCase):
    def test_chain_accepted(self):
        mcstat = gf.setup_mcmc_case_cp()
        accept = 1
        outsidebounds = 0
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        parset = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        mcstat._MCMC__update_chain(accept = accept, new_set = parset, outsidebounds = outsidebounds)
        
        self.assertTrue(np.array_equal(mcstat._MCMC__chain[-1,:], parset.theta), msg = str('theta added to end of chain - {}'.format(mcstat._MCMC__chain[-1,:])))
        self.assertEqual(mcstat._MCMC__old_set, parset, msg = 'old_set updated to new set')
        
    def test_chain_not_accepted_within_bounds(self):
        mcstat = gf.setup_mcmc_case_cp()
        accept = 0
        outsidebounds = 0
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        parset = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        mcstat._MCMC__rejected = {'total': 10, 'in_adaptation_interval': 4, 'outside_bounds': 1}
        mcstat._MCMC__update_chain(accept = accept, new_set = parset, outsidebounds = outsidebounds)
        
        self.assertTrue(np.array_equal(mcstat._MCMC__chain[-1,:], mcstat._MCMC__old_set.theta), msg = str('theta added to end of chain - {}'.format(mcstat._MCMC__chain[-1,:])))
        
# --------------------------
class UpdateRejected(unittest.TestCase):
    @classmethod
    def common_setup(cls):
        mcstat = gf.setup_mcmc_case_cp()
        rejectedin = {'total': 10, 'in_adaptation_interval': 4, 'outside_bounds': 1}
        mcstat._MCMC__rejected = rejectedin.copy()
        return mcstat, rejectedin
    
    def common_checks(self, mcstat, rejectedin, expect = 0):
        self.assertEqual(mcstat._MCMC__rejected['total'], rejectedin['total']+1, msg = 'Adds one to counter')
        self.assertEqual(mcstat._MCMC__rejected['in_adaptation_interval'], rejectedin['in_adaptation_interval']+1, msg = 'Adds one to counter')
        self.assertEqual(mcstat._MCMC__rejected['outside_bounds'], rejectedin['outside_bounds'] + expect, msg = 'This counter stays the same.')

    def test_update_rejection_stats_not_outsidebounds(self):
        mcstat, rejectedin = self.common_setup()
        mcstat._MCMC__update_rejected(outsidebounds = 0)
        self.common_checks(mcstat, rejectedin)
        
    def test_update_rejection_stats_outsidebounds(self):
        mcstat, rejectedin = self.common_setup()
        mcstat._MCMC__update_rejected(outsidebounds = 1)
        self.common_checks(mcstat = mcstat, rejectedin = rejectedin, expect = 1)
        
# --------------------------
class InitializeChain(unittest.TestCase):
    def test_initialize_chain_updatesigma_1(self):
        mcstat = gf.setup_mcmc_case_cp()
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__s2chain.shape)))
        
    def test_initialize_chain_updatesigma_0(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.simulation_options.updatesigma = 0
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain, None, msg = str('s2chain should be None -> {}'.format(mcstat._MCMC__s2chain)))
        
    def test_initialize_chain_updatesigma_1_nsos_2(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.model_settings.nsos = 2
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__s2chain.shape)))
     
# --------------------------
class ExpandChain(unittest.TestCase):
    def test_expand_chain_updatesigma_1(self):
        mcstat = gf.setup_mcmc_case_cp()
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        mcstat._MCMC__expand_chains(nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu*2 - 1, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__s2chain.shape)))
        
    def test_expand_chain_updatesigma_0(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.simulation_options.updatesigma = 0
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        mcstat._MCMC__expand_chains(nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu*2 - 1, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain, None, msg = str('s2chain should be None -> {}'.format(mcstat._MCMC__s2chain)))
        
    def test_expand_chain_updatesigma_1_nsos_2(self):
        mcstat = gf.setup_mcmc_case_cp()
        mcstat.model_settings.nsos = 2
        CL = {'theta':np.array([1.0, 2.0]), 'ss': 1.0, 'prior':0.0, 'sigma2': 0.0}
        mcstat._MCMC__old_set = ParameterSet(theta = CL['theta'], ss = CL['ss'], prior = CL['prior'], sigma2 = CL['sigma2'])
        mcstat._MCMC__chain_index = mcstat.simulation_options.nsimu - 1
        mcstat._MCMC__initialize_chains(chainind = 0, nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma, sigma2 = mcstat.model_settings.sigma2)
        mcstat._MCMC__expand_chains(nsimu = mcstat.simulation_options.nsimu, npar = mcstat.parameters.npar, nsos = mcstat.model_settings.nsos, updatesigma = mcstat.simulation_options.updatesigma)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__s2chain.shape)))
       
# --------------------------
class SetupSimulator(unittest.TestCase):
    def test_setup_simu_use_prev_false(self):
        mcstat = gf.setup_mcmc_case_cp(initialize=False)
        mcstat._MCMC__setup_simulator(use_previous_results = False)
        self.assertEqual(mcstat._MCMC__chain_index, 0, msg = 'Chain index should be 0')
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__s2chain.shape)))
        
    @patch('pymcmcstat.structures.ResultsStructure.ResultsStructure.load_json_object', return_value = {'parind': np.array([0,1,2], dtype = int), 'names': ['m', 'b', 'b2'], 'local': np.zeros([3]), 'theta': np.array([0.2, 0.5, 0.7]), 'qcov': np.array([[0.2, 0.1, 0.05],[0.1, 0.4, 0.02],[0.05, 0.02, 0.6]])})
    def test_setup_simu_use_json_file(self, mock_json_object):
        mcstat = gf.setup_mcmc_case_cp(initialize=False)
        mcstat.simulation_options.json_restart_file = 1
        mcstat._MCMC__setup_simulator(use_previous_results = False)
        self.assertTrue(np.array_equal(mcstat.simulation_options.qcov, np.array([[0.2, 0.1, 0.05],[0.1, 0.4, 0.02],[0.05, 0.02, 0.6]])), msg = 'Expect arrays to match')
        self.assertEqual(mcstat.parameters.parameters[0]['theta0'], 0.2, msg = 'Expect theta0 = 0.2')
        self.assertEqual(mcstat.parameters.parameters[1]['theta0'], -5.0, msg = 'Expect theta0 = -5.0 because sample = 0')
        self.assertEqual(mcstat.parameters.parameters[2]['theta0'], 0.7, msg = 'Expect theta0 = 0.7')
        
        self.assertEqual(mcstat._MCMC__chain_index, 0, msg = 'Chain index should be 0')
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__s2chain.shape)))
        
    def test_setup_simu_use_prev_true_causes_error(self):
        mcstat = gf.setup_mcmc_case_cp(initialize=False)
        with self.assertRaises(SystemExit, msg = 'No previous results exist'):
            mcstat._MCMC__setup_simulator(use_previous_results = True)
    
    def test_setup_simu_use_prev_true(self):
        mcstat = setup_pseudo_results(initialize = False)
        mcstat._MCMC__setup_simulator(use_previous_results = False)
        mcstat._MCMC__generate_simulation_results()
        mcstat._mcmc_status = True
        mcstat._MCMC__setup_simulator(use_previous_results = True)
        self.assertEqual(mcstat._MCMC__chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 2), msg = str('Shape should be (nsimu,2) -> {}'.format(mcstat._MCMC__chain.shape)))
        self.assertEqual(mcstat._MCMC__sschain.shape, (mcstat.simulation_options.nsimu*2 - 1, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__sschain.shape)))
        self.assertEqual(mcstat._MCMC__s2chain.shape, (mcstat.simulation_options.nsimu*2 - 1, 1), msg = str('Shape should be (nsimu,1) -> {}'.format(mcstat._MCMC__s2chain.shape)))
        
# --------------------------
class GenerateSimulationResults(unittest.TestCase):
    def test_results_generation_ntry_gt_1(self):
        mcstat = setup_pseudo_results(initialize = False)
        mcstat._MCMC__setup_simulator(use_previous_results = False)
        mcstat._MCMC__generate_simulation_results()
        results = mcstat.simulation_results.results
        self.assertTrue(mcstat.simulation_results.basic, msg = 'Basic successfully added if true')
        self.assertTrue(np.array_equal(results['R'], mcstat._covariance._R), msg = 'Arrays should match')
        self.assertTrue(np.array_equal(results['cov'], mcstat._covariance._covchain), msg = 'Arrays should match')
        check_for_these = ['simulation_options', 'model_settings', 'chain', 's2chain', 'sschain', 'drscale', 'iacce', 'RDR']
        for cft in check_for_these:
            self.assertTrue(cft in results, msg = str('{} assigned successfully'.format(cft)))
            
# ------------------------------------------------
class SaveToLogFile(unittest.TestCase):
    @patch('pymcmcstat.MCMC.MCMC._MCMC__save_chains_to_bin', return_value = None)
    def test_save_to_log_file_bin(self, mock_bin):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = True
        mcstat.simulation_options.save_to_txt = False
        savecount, lastbin = mcstat._MCMC__save_to_log_file(start = 0, end = 100)
        self.assertEqual(savecount, 0, msg = 'Expect 0')
        self.assertEqual(lastbin, 100, msg = 'Expect lastbin = end = 100')
        
    @patch('pymcmcstat.MCMC.MCMC._MCMC__save_chains_to_txt', return_value = None)
    def test_save_to_log_file_txt(self, mock_txt):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = True
        savecount, lastbin = mcstat._MCMC__save_to_log_file(start = 0, end = 100)
        self.assertEqual(savecount, 0, msg = 'Expect 0')
        self.assertEqual(lastbin, 100, msg = 'Expect lastbin = end = 100')
# --------------------------------------------------------
class SaveChainsToTXT(unittest.TestCase):
    def compare_chain(self, file, chain):
        self.assertTrue(os.path.isfile(file), msg = 'File exists')
        out = np.loadtxt(file)
        self.assertTrue(np.array_equal(out, chain), msg = str('Expect arrays to match: {} neq {}'.format(out, chain)))
        
    def test_save_chains_to_txt(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(mcstat.simulation_options, extension = 'txt')
        
        mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        self.compare_chain(chainfile, mcstat._MCMC__chain)
        self.compare_chain(sschainfile, mcstat._MCMC__sschain)
        self.compare_chain(s2chainfile, mcstat._MCMC__s2chain)
        self.compare_chain(covchainfile, np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R))
        shutil.rmtree(savedir)
        
    def test_save_chains_to_txt_updatesigma_off(self):
        mcstat = setup_case()
        mcstat.simulation_options.updatesigma = False
        mcstat._MCMC__s2chain = None
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(mcstat.simulation_options, extension = 'txt')
        
        mcstat._MCMC__save_chains_to_txt(start = 0, end = 100)
        self.compare_chain(chainfile, mcstat._MCMC__chain)
        self.compare_chain(sschainfile, mcstat._MCMC__sschain)
        self.compare_chain(covchainfile, np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R))
        shutil.rmtree(savedir)
        
# --------------------------------------------------------
class SaveChainsToBIN(unittest.TestCase):
    def compare_chain(self, file, chain):
        self.assertTrue(os.path.isfile(file), msg = 'File exists')
        out = ChainProcessing.read_in_bin_file(file)
        self.assertTrue(np.array_equal(out, chain), msg = str('Expect arrays to match: {}'.format(file)))
        
    def test_save_chains_to_bin(self):
        mcstat = setup_case()
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(mcstat.simulation_options, extension = 'h5')
        
        mcstat._MCMC__save_chains_to_bin(start = 0, end = 100)
        self.compare_chain(chainfile, mcstat._MCMC__chain)
        self.compare_chain(sschainfile, mcstat._MCMC__sschain)
        self.compare_chain(s2chainfile, mcstat._MCMC__s2chain)
        self.compare_chain(covchainfile, np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R))
        shutil.rmtree(savedir)
        
    def test_save_chains_to_bin_updatesigma_off(self):
        mcstat = setup_case()
        mcstat.simulation_options.updatesigma = False
        mcstat._MCMC__s2chain = None
        savedir = gf.generate_temp_folder()
        mcstat.simulation_options.savedir = savedir
        
        chainfile, s2chainfile, sschainfile, covchainfile = ChainProcessing._create_path_with_extension_for_all_logs(mcstat.simulation_options, extension = 'h5')
        
        mcstat._MCMC__save_chains_to_bin(start = 0, end = 100)
        self.compare_chain(chainfile, mcstat._MCMC__chain)
        self.compare_chain(sschainfile, mcstat._MCMC__sschain)
        self.compare_chain(covchainfile, np.dot(mcstat._covariance._R.transpose(),mcstat._covariance._R))
        shutil.rmtree(savedir)
        
# --------------------------------------------------------
class RunSimulation(unittest.TestCase):
    def test_run_simulation(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.nsimu = 100
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = False
        mcstat.run_simulation()
        self.assertTrue(mcstat._mcmc_status, msg = 'Expect True if successfully run')
        
        check_these = ['mcmcplot', 'PI', 'chainstats']
        for ci in check_these:
            self.assertTrue(hasattr(mcstat, ci), msg = str('object has attribute: {}'.format(ci)))
            
    def test_run_simulation_with_json(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.nsimu = 100
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = False
        tmpfile = gf.generate_temp_file(extension = 'json')
        tmpfolder = gf.generate_temp_folder()
        os.mkdir(tmpfolder)
        mcstat.simulation_options.savedir = tmpfolder
        mcstat.simulation_options.results_filename = tmpfile
        mcstat.simulation_options.save_to_json = True
        mcstat.run_simulation()
        self.assertTrue(mcstat._mcmc_status, msg = 'Expect True if successfully run')
        
        check_these = ['mcmcplot', 'PI', 'chainstats']
        for ci in check_these:
            self.assertTrue(hasattr(mcstat, ci), msg = str('object has attribute: {}'.format(ci)))
            
        shutil.rmtree(tmpfolder)
        
    def test_run_simulation_verbose(self):
        mcstat = gf.basic_mcmc()
        mcstat.simulation_options.nsimu = 200
        mcstat.simulation_options.save_to_bin = False
        mcstat.simulation_options.save_to_txt = False
        mcstat.simulation_options.verbosity = 20
        mcstat.run_simulation()
        self.assertTrue(mcstat._mcmc_status, msg = 'Expect True if successfully run')
        
        check_these = ['mcmcplot', 'PI', 'chainstats']
        for ci in check_these:
            self.assertTrue(hasattr(mcstat, ci), msg = str('object has attribute: {}'.format(ci)))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 26 06:17:24 2018

@author: prmiles
"""
from pymcmcstat.utilities.general import message, removekey
import unittest
import io
import sys

# --------------------------
class MessageDisplay(unittest.TestCase):
    def test_standard_print(self):
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        flag = message(verbosity = 1, level = 0, printthis = 'test')
        sys.stdout = sys.__stdout__                     # Reset redirect.
        self.assertEqual(capturedOutput.getvalue(), 'test\n', msg = 'Expected string')
        self.assertTrue(flag, msg = 'Statement was printed')
        
    def test_no_print(self):
        capturedOutput = io.StringIO()                  # Create StringIO object
        sys.stdout = capturedOutput                     #  and redirect stdout.
        flag = message(verbosity = 0, level = 1, printthis = 'test')
        sys.stdout = sys.__stdout__                     # Reset redirect.
        self.assertFalse(flag, msg = 'Statement not printed because verbosity less than level')
        
# --------------------------
class RemoveKey(unittest.TestCase):
    def test_removekey(self):
        d = {'a1': 0.1, 'a2': 0.2, 'a3': 'hello'}
        r = removekey(d, 'a2')
        check_these = ['a1', 'a3']
        for ct in check_these:
            self.assertEqual(d[ct], r[ct], msg = str('Expect element agreement: {}'.format(ct)))
            
        self.assertFalse('a2' in r, msg = 'Expect removal')